// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "event.h"
#include "sweepobject.h"

using namespace vorosweep;

const char *Event::eventtypename[] =
{
  "NEWGENERATOR",
  "NEWEDGESWITCH",
  "NEWBORDEREDGESWITCH",
  "NEWFACETSWITCH",
  "MULTICRASH",
  "EDGE_EDGECRASH1",
  "EDGE_EDGECRASH2",
  "EDGE_FACETCRASH",
  "EDGE_BORDERCRASH",
  "FACET_BORDERCRASH",
  "EDGESWITCH",
  "FACETSWITCH",
  "BORDEREDGESWITCH"
};

bool EventPtrComp::operator() ( const Event* a, const Event* b )
{
// 	printf("%p : a pos %lf %lf time : %.20lf, %p : b pos %lf %lf time : %.20lf --> %d\n", a, a->p[0], a->p[1], a->time, b, b->p[0], b->p[1], b->time, a->time == b->time);
  if ( a == b )
    return false;
  else if ( a->time < b->time )
    return true;
  // in case the time is the same as well as the sweepobject, then check the distance
  else if ( a->time == b->time && a->dist < b->dist )
    return true;
  else if ( a->time == b->time && a->dist == b->dist && a->so[0] < b->so[0] )
    return true;
  else if ( a->time == b->time && a->dist == b->dist && a->so[0] == b->so[0] && a->so[1] < b->so[1] )
    return true;
  else if ( a->time == b->time && a->dist == b->dist && a->so[0] == b->so[0] && a->so[1] == b->so[1] && a->gen < b->gen )
    return true;
  else
    return false;
}


Event::Event ()
{
  so[0] = so[1] = NULL;
  brd = NULL;
  gen = NULL;
  p[0] = p[1] = 0.0;
  time = 0.0;
  dist = 0.0;
  status = Event::ENABLE;
  gid = -1;
	dist = INFINITY;
  init_extras ( );
//   ambiguous = 0;
// 	priority = Event::LOW;
}

void Event::init ( SweepObject *s, double *pos, double t, eventtype what, int index_in_grid )
{
  so[0] = s;
  so[1] = NULL;
  brd = NULL;
  gen = NULL;
  p[0] = pos[0];
  p[1] = pos[1];
  time = t;
//   dist = geom::sqrtdistance2d ( so[0]->get_start()->p, pos );
  type = what;
  gid = index_in_grid;
	init_dist ( );
  init_extras ( );

  // everything is done before insertion
  so[0]->insert_event ( this );

  dbgprintf ( 1, "new : " );
  print();
}

void Event::init ( SweepObject *s0, SweepObject *s1, double *pos, double t, eventtype what, int index_in_grid )
{
  so[0] = s0;
  so[1] = s1;
  brd = NULL;
  gen = NULL;
  p[0] = pos[0];
  p[1] = pos[1];
  time = t;
  type = what;
  gid = index_in_grid;
	init_dist ( );

  init_extras ( );

  // everything is done before insertion
  so[0]->insert_event ( this );
  so[1]->insert_event ( this );

  dbgprintf ( 1, "new : " );
  print();
//   getchar();
}

void Event::init ( SweepObject *s, Border *b, double *pos, double t, eventtype what, int index_in_grid )
{
  so[0] = s;
  so[1] = NULL;
  brd = b;
  gen = NULL;
  p[0] = pos[0];
  p[1] = pos[1];
  time = t;
  type = what;
  gid = index_in_grid;
	init_dist ( );
  init_extras ( );

  // everything is done before insertion
  so[0]->insert_event ( this );

  dbgprintf ( 1, "new : " );
  print();
//   getchar();
}

void Event::init ( Generator *g, double t, eventtype what )
{
  so[0] = NULL;
  so[1] = NULL;
  brd = NULL;
  gen = g;
  p[0] = p[1] = 0.0;
  time = t;
  dist = INFINITY;
  type = what;
  gid = -1;
  extras.clear();
//   ambiguous = 0;
  dbgprintf ( 1, "new : " );
  print();
//   getchar();
}

SweepObject *Event::get_sweepobject ( int i )
{
  return so[i];
}
SweepEdge *Event::get_sweepedge ( int i )
{
  return reinterpret_cast<SweepEdge*> ( so[i] );
}
SweepFacet *Event::get_sweepfacet ( int i )
{
  return reinterpret_cast<SweepFacet*> ( so[i] );
}
// Border *Event::get_border ()
// {
//   return reinterpret_cast<Border*> ( so[1] );
// }
// Generator *Event::get_generator ()
// {
//   return reinterpret_cast<Generator*> ( so[0] );
// }
void Event::init_dist ( )
{
	dist = INFINITY;
  if ( so[0]->get_xyspeed ( ) == INFINITY )
    dist = geom::sqrtdistance2d ( so[0]->get_start()->p, p );
  if ( so[1] && so[1]->get_xyspeed ( ) == INFINITY )
  {
    double tmp = geom::sqrtdistance2d ( so[1]->get_start()->p, p );
    if ( tmp < dist )
      dist = tmp;
  }
}
void Event::init_extras ( )
{
//   dbgprintf ( 2, "== init extra edges ==\n" );
  extras.clear();
  for ( int i = 0; i < 2; i++ )
  {
    if ( so[i] && ( so[i]->get_type() == SweepObject::EDGE || so[i]->get_type() == SweepObject::BORDEREDGE ) )
    {
      dbgprintf ( 2, "init with edge %p\n", so[i] );
      extras.push_back ( reinterpret_cast<SweepEdge*> ( so[i] ) );
    }
  }
//   dbgprintf ( 2, "--> now %d extra edges\n", ( int ) extras.size() );
}
void Event::append_extra ( SweepEdge *se )
{
  dbgprintf ( 2, "--> appending extra edges among %d\n", ( int ) extras.size() );

  // check
  for ( int i = 0; i < ( int ) extras.size(); i++ )
    if ( se == extras[i] )
      return;

  if ( type != Event::MULTICRASH )
  {
    dbgprintf ( 2, "--> converting event\n" );
    type = Event::MULTICRASH;
  }

  // append
  dbgprintf ( 2, "append edge %p\n", se );
  extras.push_back ( se );
  dbgprintf ( 2, "--> now %d extra edges\n", ( int ) extras.size() );

  se->insert_event ( this );
}
void Event::print()
{
  dbgprintf ( 1, "event %p type %s", this, eventtypename[get_type() ] );
  if ( so[0] )
  {
    dbgprintf ( 1, " with object type %s : ", SweepObject::objectname[so[0]->get_type() ] );
//     dbgprintf ( 1, "pos %lf %lf, dir %lf %lf, time %.10lf, dist %.10lf, bucket %d\n", p[0], p[1], so[0]->get_dir3d() [0], so[0]->get_dir3d() [1], time, dist, gid );
    dbgprintf ( 1, "pos %lf %lf, time %.10lf, dist %.10lf, bucket %d\n", p[0], p[1], time, dist, gid );
  }
  else
  {
    dbgprintf ( 1, " with object type GENERATOR : pos %lf %lf, time %.20lf, bucket %d\n", p[0], p[1], time, gid );
  }
}


