// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "common.h"
#include "parameters.h"
#include "elementary.h"
#include <geom.h>

#ifndef _EVENT_INCLUDE
#define _EVENT_INCLUDE

namespace vorosweep
{
class Border;
class SweepObject;
class SweepEdge;
class SweepFacet;
class Generator;

class Event
{
  friend class EventPtrComp;

public:
  static const char *eventtypename[];
public:

  enum eventtype
  {
    NEWGENERATOR,
    NEWEDGESWITCH,
    NEWBORDEREDGESWITCH,
    NEWFACETSWITCH,
    MULTICRASH,
    EDGE_EDGECRASH1,
    EDGE_EDGECRASH2,
    EDGE_FACETCRASH,
    EDGE_BORDERCRASH,
    FACET_BORDERCRASH,
    EDGESWITCH,
    FACETSWITCH,
    BORDEREDGESWITCH
  };

  enum eventstatus
  {
    ENABLE,
    DISABLE
  };
// 		enum eventpriority
//     {
// 			LOW,
// 			HIGH
// 		};
private:
  double p[2];
  double time; // time of event
  double dist; // dist of event
  eventtype type;
  eventstatus status;
  int gid;
// 	int ambiguous;
  // the following things should be unioned :
  SweepObject *so[2];
  Border *brd;
  Generator *gen;
  std::vector<SweepEdge *> extras;
public:
  std::set<Event *>::iterator it;

public:
  Event ();
  void init ( SweepObject *s, double *pos, double t, eventtype what, int index_in_grid = -1 );
  void init ( SweepObject *s0, SweepObject *s1, double *pos, double t, eventtype what, int index_in_grid = -1 );
  void init ( SweepObject *s, Border *b, double *pos, double t, eventtype what, int index_in_grid = -1 );
  void init ( Generator *g, double t, eventtype what );
  void init_dist ( );
  void init_extras();
  void append_extra ( SweepEdge *se );
  inline std::vector<SweepEdge*>& get_extras()
  {
    return extras;
  }
  SweepObject *get_sweepobject ( int i = 0 );
  SweepEdge *get_sweepedge ( int i = 0 );
  SweepFacet *get_sweepfacet ( int i = 0 );
// 		Border *get_border ();
// 		Generator *get_generator ();
  inline Border *get_border ()
  {
    return brd;
  }
  inline Generator *get_generator ()
  {
    return gen;
  }
  inline double *get_coord()
  {
    return p;
  }
  inline double get_time()
  {
    return time;
  }
  inline double get_dist()
  {
    return dist;
  }
  inline void update ( double t, double d ) /// be very careful with this
  {
    time = t;
//     if ( d < dist )
      dist = d;
  }
  inline eventtype get_type()
  {
    return type;
  }
  inline int get_gid()
  {
    return gid;
  }
  inline void disable()
  {
    dbgprintf ( 2, "disabling event %p\n", this );
    status = Event::DISABLE;
  }
  inline int enabled()
  {
    return ( status == Event::ENABLE );
  }
  inline int disabled()
  {
    return ( status == Event::DISABLE );
  }
  void print();
};

class EventPtrComp
{
public:
  bool operator() ( const Event* a, const Event* b );
// 	private:
// 		bool comp_facetswitch(const Event* a, const Event* b);
};

// class BorderEvent : public Event
// {
//     private:
//       Border *brd;
//     public:
//     void init ( SweepObject *s, Border *b, double *pos, double t, eventtype what, int index_in_grid );
//     inline Border *get_border ()
//     {
//       return brd;
//     }
// };

};

#endif


