// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"
// #include "color.h"
#include "seidel/interface.h"

using namespace vorosweep;

void Graph::export_asymptote ( const char *name )
{
  if ( ! ( int ) pointlist.size() || ! ( int ) indexmap.size() )
  {
    printf ( "--> You have to call generate_internals before exporting\n" );
    return;
  }

  char filename[100];
  static int count = 0;
  sprintf ( filename, "%s_%d.asy", name, count++ );
  tprintf ( "Writing file %s ...\n", filename );
  FILE *fp = fopen ( filename, "w" );

  int nbf = 0;
  int nbt = 0;

//   double eye[3] = {1.5,-1.0,6.0};
//   double look[3] = {0.5,0.5,0.5};
  double look[3] = {1.0,-1.0,7.0};

  fprintf ( fp, "settings.outformat=\"pdf\";\n" );
  fprintf ( fp, "settings.prc=false;\n" );
  fprintf ( fp, "settings.render=8;\n" );
  fprintf ( fp, "import three;\n" );
  fprintf ( fp, "import solids;\n" );
  fprintf ( fp, "size(10cm,0);\n" );
  fprintf ( fp, "currentprojection=orthographic((%lf, %lf, %lf),up=Z);\n", look[0], look[1], look[2] );
// currentlight=nolight;
  fprintf ( fp, "viewportmargin=(10mm,10mm);\n" );

  for ( int i = 0; i < ( int ) generatorlist.size(); i++ )
  {
    Generator *g = generatorlist[i];
    if ( g->is_active() )
    {
//       // draw center
//       board.setPenColor ( LibBoard::Color::Black );
//       board.fillCircle ( cg->get_apex()->p[0], cg->get_apex()->p[1], 0.002, 0 );

      // draw voronoi edges
      for ( int j = g->face_size()-1; j >= 0; j-- )
      {
        SweepFacet *sf = g->get_sweepfacet ( j );
        FrontLine *fl = sf->get_frontline();
        std::list<Polygon *> plist = fl->get_polygons();
        for ( std::list<Polygon *>::iterator itp = plist.begin(); itp != plist.end(); itp++ )
        {
          Polygon *pol = *itp;
          std::list<SweepEdge *> elist = pol->get_edges();
          std::list<SweepEdge *>::reverse_iterator it = elist.rbegin();

          fprintf ( fp, "path3 f%d = ", nbf++ );

          double vx[1000][2];
          double vxz[1000];
          int nbv = 0;
          Point *prev = NULL;
          for ( it = elist.rbegin(); it != elist.rend(); it++ )
          {
            SweepEdge *se = ( *it );
            Point *pt = NULL;
            if ( se->get_sweepfacet ( 1 ) == sf )
              pt = se->get_current();
            else
              pt = se->get_start();

            if ( !prev || geom::distance3d ( pt->p, prev->p ) > Parameters::instance()->vorosweep.epsilon )
            {
              vx[nbv+1][0] = pt->p[0];
              vx[nbv+1][1] = pt->p[1];
              vxz[nbv+1] = pt->p[2];
              nbv++;
            }
            prev = pt;
          }

          for ( int i = 1; i <= nbv; i++ )
          {
            fprintf ( fp, "(%lf,%lf,%lf) -- ", vx[i][0], vx[i][1], vxz[i] );
//             printf ( "(%lf,%lf,%lf) -- ", vx[i][0], vx[i][1], vxz[i] );
          }

          fprintf ( fp, "cycle;\n" );
//           printf ( "cycle;\n" );

          int ncontours = 1; // #contours
          int cntr[1]; // An array describing the number of points in each contour.
          int triangles[300][3];
          cntr[0] = nbv;
          triangulate_polygon ( ncontours, cntr, vx, triangles ); // The number of output triangles produced for a polygon with n points is (n - 2) + 2*(#holes)

          for ( int i = 0; i < nbv-2; i++ )
          {
            int id0 = triangles[i][0];
            int id1 = triangles[i][1];
            int id2 = triangles[i][2];

            fprintf ( fp, "path3 t%d = (%lf, %lf, %lf) -- (%lf, %lf, %lf) -- (%lf, %lf, %lf) -- cycle;\n", nbt,
                      vx[id0][0], vx[id0][1], vxz[id0],
                      vx[id1][0], vx[id1][1], vxz[id1],
                      vx[id2][0], vx[id2][1], vxz[id2]
                    );
            fprintf ( fp, "draw(surface(t%d),gray+opacity(.4),0bp+gray);\n", nbt );
            nbt++;
          }

//           for ( int i = 1; i <= nbv; i++ )
//           {
//             double localeps = 1e-3;
//             fprintf ( fp, "	def pp%d [[hp]]*(%lf,%lf,%lf)\n", i, vx[i][0] + localeps* ( eye[0]-look[0] ), vx[i][1] + localeps* ( eye[1]-look[1] ), vxz[i] + localeps* ( eye[2]-look[2] ) );
//             if ( i>=2 )
//             {
//               fprintf ( fp, "	line[linecolor=black,linewidth=1.0pt](pp%d)(pp%d)\n", i-1, i );
//             }
//             fprintf ( fp, "	def ppp%d [[hp]]*(%lf,%lf,%lf)\n", i, vx[i][0] + 2.0*localeps* ( eye[0]-look[0] ), vx[i][1] + 2.0*localeps* ( eye[1]-look[1] ), vxz[i] + 2.0*localeps* ( eye[2]-look[2] ) );
//             if ( i>=2 )
//             {
//               fprintf ( fp, "	line[lay=over,linecolor=black,linewidth=1.0pt,linestyle=dashed](ppp%d)(ppp%d)\n", i-1, i );
//             }
//           }
        }
      }
    }
  }

  for ( int i = 0; i < nbf; i++ )
  {
    fprintf ( fp, "draw(f%d,1bp+black);\n", i );
  }

//   fprintf ( fp, "def axes {\n" );
//   fprintf ( fp, "	def O (0,0,0)\n" );
//   fprintf ( fp, "	def sz 1\n" );
//   fprintf ( fp, "	line [lay=over,arrows=->] (O)(sz,0,0)\n" );
//   fprintf ( fp, "	line [lay=over,arrows=->] (O)(0,sz,0)\n" );
//   fprintf ( fp, "	line [lay=over,arrows=->] (O)(0,0,sz)\n" );
//   fprintf ( fp, "	special |\\uput[r]#1{$x$}\\uput[u]#2{$y$}\\uput[l]#3{$z$}|(sz,0,0)(0,sz,0)(0,0,sz)\n" );
//   fprintf ( fp, "}\n" );
//
//   fprintf ( fp, "def eye (%lf, %lf, %lf)\n", eye[0], eye[1], eye[2] );
//   fprintf ( fp, "def look_at (%lf, %lf, %lf)\n", look[0], look[1], look[2] );
//   fprintf ( fp, "put { view((eye), (look_at)) } { {axes} " );
//   for ( int i = 0; i < nbf; i++ )
//   {
//     fprintf ( fp, "{f%d}", i );
//   }
//   fprintf ( fp, " }\n" );

  fclose ( fp );

  tprintf ( "Done\n" );
}
