// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "generator.h"

using namespace vorosweep;

double stable_signedpow ( double num, double exponent )
{
  double result;
  if ( fabs ( num ) > 1e-10 )
    result = copysign ( pow ( fabs ( num ), exponent ), num );
  else
    result = 0.0;
//   printf ("num = %.10lf, pow = %lf\n", num, result);
  return result;
}

double stable_unsignedpow ( double num, double exponent )
{
  double result;
  if ( fabs ( num ) > 1e-10 )
    result = pow ( fabs ( num ), exponent );
  else
    result = 0.0;
//   printf ("num = %.10lf, pow = %lf\n", num, result);
  return result;
}

// // Generator::Generator ( double t0, int nface )
// // {
// // 	starttime = t0;
// //   nf = nface;
// // 	ise.resize ( nf );
// //   isf.resize ( nf );
// //   se.resize ( nf );
// //   sf.resize ( nf );
// // 	for (int i = 0; i < nf; i++)
// // 	{
// // 		ise[i] = new double[3];
// // 		isf[i] = new double[3];
// // 	}
// //   state = Generator::NOACTIVE;
// // }
// //
// // Convex_generator::Convex_generator ( double *pt, double t0, int nface, double a, double v1, double v2, int index )
// // {
// //   apex = basic_type::pcontainer->get();
// //   apex->set ( pt );
// //   int id = basic_type::pcontainer->get_index();
// //   apex->set_index ( id );
// //   apex->p[2] = t0;
// //   angle = a;
// //   v[0] = v1;
// //   v[1] = v2;
// //   type = Generator::POINT;
// // }

void Generator::compute_normal_plane ( double p, double t, double *plane )
{
  double vec[2];
  vec[0] = cos ( t );
  vec[1] = sin ( t );
  double ntmp[3];

  compute_normal_plane ( p, vec, ntmp );

  plane[0] = cos ( A->angle ) *ntmp[0] - sin ( A->angle ) *ntmp[1];
  plane[1] = cos ( A->angle ) *ntmp[1] + sin ( A->angle ) *ntmp[0];
  plane[2] = ntmp[2];
}

void Generator::compute_normal_plane ( double p, double *vec, double *plane )
{
  double s = 1.0;
  double tgttmp[3];
  double tgstmp[3];
  double ntmp[3];

  if ( fabs ( vec[0] ) < 1e-10 )
  {
//     tgttmp[0] = copysign(1.0, sint);
    tgttmp[0] = - ( 2.0*A->v[0]*s*vec[1] ) /p;
    tgttmp[1] = 0.0;
  }
  else if ( fabs ( vec[1] ) < 1e-10 )
  {
    tgttmp[0] = 0.0;
//     tgttmp[1] = copysign(1.0, cost);
    tgttmp[1] = ( 2.0*A->v[1]*s*vec[0] ) /p;
  }
  else
  {
    tgttmp[0] = - ( 2.0*A->v[0]*s*vec[1]*stable_unsignedpow ( vec[0], 2.0/p-1.0 ) ) /p;
    tgttmp[1] = ( 2.0*A->v[1]*s*vec[0]*stable_unsignedpow ( vec[1], 2.0/p-1.0 ) ) /p;
  }
  tgttmp[2] = 0.0;

  tgstmp[0] = A->v[0]*stable_signedpow ( vec[0], 2.0/p );
  tgstmp[1] = A->v[1]*stable_signedpow ( vec[1], 2.0/p );
  tgstmp[2] = 1.0;

  ntmp[0] = tgttmp[1]*tgstmp[2]-tgttmp[2]*tgstmp[1];
  ntmp[1] = tgttmp[2]*tgstmp[0]-tgttmp[0]*tgstmp[2];
  ntmp[2] = tgttmp[0]*tgstmp[1]-tgttmp[1]*tgstmp[0];

  geom::copyvec3d ( ntmp, plane );
//   plane[0] = cos ( A->angle ) *ntmp[0] - sin ( A->angle ) *ntmp[1];
//   plane[1] = cos ( A->angle ) *ntmp[1] + sin ( A->angle ) *ntmp[0];
//   plane[2] = ntmp[2];
}

void Generator::generate_radial_subdivision ( )
{
  double asub = 2.0*M_PI / ( double ) cone->size(); // angle subdivision

  dbgprintf ( 2, "nf --> %d, angle subd --> %lf\n", cone->size(), asub );

  if ( Parameters::instance()->vorosweep.perturb_angles )
    angle += 1e-5*drand48();

  A = new matrix ( angle, v[0], v[1] );

	
	order = Parameters::instance()->generator.order;
	if (order < 1.0) // random
		order = (double)(2*(rand() % 10) + 2);

  for ( int i = 0; i < cone->size(); i++ )
  {
    double t = i* asub;

    dbgprintf ( 2, "t --> %lf\n", t );

    double plane[3];
    compute_normal_plane ( order, t, plane );
    geom::normalize3d ( plane );

    double tangent[3];
    tangent[0] = plane[1];
    tangent[1] = -plane[0];
    tangent[2] = 0.0;

    double dir[3];
    geom::cross_product3d ( tangent, plane, dir );
    geom::normalize3d ( dir );

//     printf ( "dir --> %lf %lf\n", dir[0], dir[1] );
    cone->set_face ( i, plane, dir );
  }

  cone->generate_edges();
}


void Generator::update_current ( double time )
{
  if ( !is_active() )
    return;
  for ( int i = 0; i < face_size(); i++ )
  {
    sf[i]->update_current ( time );
    FrontLine *fl = sf[i]->get_frontline();
    fl->update_current ( time );
  }
}

int Generator::find_interval ( double *newplane, int *index, double *newdir0, double *newdir1 )
{
//   int index[2];
  double *plane;
  double *se0dir;
  double *se1dir;
  double newdir[2][3];
  index[0] = 0;
  int j = 0;

  printf ( "newplane --> %lf %lf %lf\n", newplane[0], newplane[1], newplane[2] );

  for ( int i = 0; ; i++ )
  {
    index[j] = get_next_index ( index[j] );
    {
      plane = get_normal ( index[j] );
      printf ( "plane %d --> %lf %lf %lf\n", j, plane[0], plane[1], plane[2] );
      se0dir = get_edge ( index[j], 0 );
      se1dir = get_edge ( index[j], 1 );
      if ( !j )
        geom::cross_product3d ( plane, newplane, newdir[j] );
      else
        geom::cross_product3d ( newplane, plane, newdir[j] );
      geom::normalize3d ( newdir[j] );
      printf ( "newdir --> %lf %lf %lf\n", newdir[j][0], newdir[j][1], newdir[j][2] );
//       printf ( "%lf %lf <-- --> %lf %lf\n", se0dir[j][0], se0dir[j][1], se1dir[j][0], se1dir[j][1] );
    }

		if ( newdir[j][2] > 0.0 && geom::vector_is_between ( newdir[j], se0dir, se1dir ))
    {
			printf("ok !!\n");
      if ( j == 0 )
      {
        j++;
        index[1] = index[0];
      }
      else
      {
        geom::copyvec3d ( newdir[0], newdir0 );
        geom::copyvec3d ( newdir[1], newdir1 );
        return 1;
      }
    }

  }
  return 0;
}


void Convex_generator::generate ( )
{
  generate_radial_subdivision( );
  for ( int i = 0; i < cone->size(); i++ )
  {
    se[i] = new SweepEdge ( apex, get_start_time() );
    se[i]->set_init ();
  }

  for ( int i = 0; i < cone->size(); i++ )
  {
    double *plane = cone->get_normal ( i );
    double *dir = cone->get_dir ( i );

    dbgprintf ( 2, "plane %d --> %lf %lf %lf\n", i, plane[0], plane[1], plane[2] );
    dbgprintf ( 2, "dir %d --> %lf %lf %lf\n", i, dir[0], dir[1], dir[2] );
    sf[i] = new SweepFacet ( apex, plane, dir, get_start_time() );
    sf[i]->set_pos ( i );
    double xys = sf[i]->compute_squared_speeds ( dir );
    sf[i]->set_xyspeed ( sqrt ( xys ) );
    sf[i]->set_generator ( this );
    /*FrontLine *fl = */
    new FrontLine ( sf[i], se[i], se[ ( i + ( cone->size() - 1 ) ) % cone->size() ] );
  }

  for ( int i = 0; i < cone->size(); i++ )
  {
    double *edge = cone->get_edge ( i );
    dbgprintf ( 2, "edge %d dir --> %lf %lf %lf\n", i, edge[0], edge[1], edge[2] );
    se[i]->set_dir ( edge );
  }

  for ( int i = 0; i < cone->size(); i++ )
  {
    SweepFacet *sf = get_sweepfacet ( i );
    SweepEdge *se = get_sweepedge ( i );
    se->update_speeds ( sf->get_plane() );
  }
//     getchar();
}

void Linear_generator::generate ( )
{
  generate_radial_subdivision( );

  double line[2];
  geom::vec2d ( apex[0]->p, apex[1]->p, line );
  geom::normalize2d ( line );
  double ortho[2][2];
  ortho[0][0] = -line[1];
  ortho[0][1] = line[0];
  ortho[1][0] = line[1];
  ortho[1][1] = -line[0];
  double orthoplane[2][3];
  compute_normal_plane ( 2.0, ortho[0], orthoplane[0] );
  compute_normal_plane ( 2.0, ortho[1], orthoplane[1] );

  printf ( "apex --> line %lf %lf\n", line[0], line[1] );

  int index[2][2];
  double *plane;
  double *dir;
//   double *se0dir;
//   double *se1dir;
  double newdir0[3];
  double newdir1[3];


  find_interval ( orthoplane[0], index[0], newdir0, newdir1 );
  printf ( "index --> %d %d\n", index[0][0], index[0][1] );
  find_interval ( orthoplane[1], index[1], newdir0, newdir1 );
  printf ( "index --> %d %d\n", index[1][0], index[1][1] );

  int size = cone->size() - ( index[0][1]-index[0][0]-1 ) - ( index[1][1]-index[1][0]-1 ) + 2;
  se.resize ( size );
  sf.resize ( size );

  printf ( "size --> %d\n", size );

  int count = 0;
  for ( int i = index[0][1]; ; )
  {
    printf ( "init sweepedge %d\n", count );
    se[count] = new SweepEdge ( apex[0], get_start_time() );
    printf ( "current --> %p\n", se[count]->get_current() );
    se[count]->set_init ();
    count++;
    if ( i == index[1][0] )
    {
      se[count] = new SweepEdge ( apex[0], get_start_time() );
      printf ( "current --> %p\n", se[count]->get_current() );
      se[count]->set_init ();
      count++;
      break;
    }
    i = ( i+1 ) %cone->size();
  }
  for ( int i = index[1][1]; ; )
  {
    printf ( "init sweepedge %d\n", count );
    se[count] = new SweepEdge ( apex[1], get_start_time() );
    printf ( "current --> %p\n", se[count]->get_current() );
    se[count]->set_init ();
    count++;
    if ( i == index[0][0] )
    {
      se[count] = new SweepEdge ( apex[1], get_start_time() );
      printf ( "current --> %p\n", se[count]->get_current() );
      se[count]->set_init ();
      count++;
      break;
    }
    i = ( i+1 ) %cone->size();
  }

  printf ( "--> init sweepfacets\n" );

  count = 0;
	double ndir[3];
  geom::dir2d_to_dir3d ( ortho[0], orthoplane[0], ndir );
//   dbgprintf ( 2, "plane %d --> %lf %lf %lf\n", count, plane[0], plane[1], plane[2] );
  dbgprintf ( 2, "dir %d --> %lf %lf %lf\n", count, ndir[0], ndir[1], ndir[2] );
  sf[count] = new SweepFacet ( apex[0], orthoplane[0], ndir, get_start_time() );
  sf[count]->set_pos ( index[0][0], 1 );
  sf[count]->set_pos ( index[0][1], 0 );
  double xys = sf[count]->compute_squared_speeds ( ndir );
  sf[count]->set_xyspeed ( sqrt ( xys ) );
  sf[count]->set_generator ( this );
  /*FrontLine *fl = */
  new FrontLine ( sf[count], se[count], se[ ( count + ( size - 1 ) ) % size ] );
  count++;

  for ( int i = index[0][1]; ; )
  {
    plane = cone->get_normal ( i );
    dir = cone->get_dir ( i );
    dbgprintf ( 2, "plane %d --> %lf %lf %lf\n", count, plane[0], plane[1], plane[2] );
    dbgprintf ( 2, "dir %d --> %lf %lf %lf\n", count, dir[0], dir[1], dir[2] );
    sf[count] = new SweepFacet ( apex[0], plane, dir, get_start_time() );
    sf[count]->set_pos ( i );
    xys = sf[count]->compute_squared_speeds ( dir );
    sf[count]->set_xyspeed ( sqrt ( xys ) );
    sf[count]->set_generator ( this );
    /*FrontLine *fl = */
    new FrontLine ( sf[count], se[count], se[ ( count + ( size - 1 ) ) % size ] );
    count++;
    if ( i == index[1][0] )
    {
      break;
    }
    i = ( i+1 ) %cone->size();
  }

  geom::dir2d_to_dir3d ( ortho[1], orthoplane[1], ndir );
//   dbgprintf ( 2, "plane %d --> %lf %lf %lf\n", count, plane[0], plane[1], plane[2] );
  dbgprintf ( 2, "dir %d --> %lf %lf %lf\n", count, ndir[0], ndir[1], ndir[2] );
  sf[count] = new SweepFacet ( apex[1], orthoplane[1], ndir, get_start_time() );
  sf[count]->set_pos ( index[1][0], 1 );
  sf[count]->set_pos ( index[1][1], 0 );
  xys = sf[count]->compute_squared_speeds ( ndir );
  sf[count]->set_xyspeed ( sqrt ( xys ) );
  sf[count]->set_generator ( this );
  /*FrontLine *fl = */
  new FrontLine ( sf[count], se[count], se[ ( count + ( size - 1 ) ) % size ] );
  count++;

  for ( int i = index[1][1]; ; )
  {
    plane = cone->get_normal ( i );
    dir = cone->get_dir ( i );
    dbgprintf ( 2, "plane %d --> %lf %lf %lf\n", count, plane[0], plane[1], plane[2] );
    dbgprintf ( 2, "dir %d --> %lf %lf %lf\n", count, dir[0], dir[1], dir[2] );
    sf[count] = new SweepFacet ( apex[1], plane, dir, get_start_time() );
    sf[count]->set_pos ( i );
    xys = sf[count]->compute_squared_speeds ( dir );
    sf[count]->set_xyspeed ( sqrt ( xys ) );
    sf[count]->set_generator ( this );
    /*FrontLine *fl = */
    new FrontLine ( sf[count], se[count], se[ ( count + ( size - 1 ) ) % size ] );
    count++;
    if ( i == index[0][0] )
    {
      break;
    }
    i = ( i+1 ) %cone->size();
  }


  for ( int j = 0; j < size; j++ )
  {
    double edge[3];
    geom::cross_product3d ( sf[j]->get_plane(), sf[ ( j+1 ) %size]->get_plane(), edge );
    dbgprintf ( 2, "edge %d dir --> %lf %lf %lf\n", j, edge[0], edge[1], edge[2] );
    se[j]->set_dir ( edge );
  }

  for ( int j = 0; j < size; j++ )
  {
    SweepFacet *sf = get_sweepfacet ( j );
    SweepEdge *se = get_sweepedge ( j );
    se->update_speeds ( sf->get_plane() );
  }
//     getchar();
}

