// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "grid.h"
#include "border.h"
#include "event.h"

using namespace vorosweep;

Parameters *basic_grid::pm = NULL;

void Grid1d::print()
{
  dbgprintf ( 1, "## grid : start %lf %lf, end %lf %lf\n", pt[0]->p[0], pt[0]->p[1], pt[1]->p[0], pt[1]->p[1] );
  dbgprintf ( 1, "## --> %d ( dx %lf )\n## ==>", nbx, blx );
  for ( int i = 0; i < nbx+1; i++ )
    dbgprintf ( 1, " (%lf, %lf)", pt[0]->p[0] + i*blx*dir[0], pt[0]->p[1] + i*blx*dir[1] );
  dbgprintf ( 1, "\n" );
  dbgprintf ( 1, "## dir --> %lf %lf\n", dir[0], dir[1] );
}

void Grid2d::init ( int nbx, int nby )
{
  bb->lock();
  nb[0] = nbx;
  nb[1] = nby;
  objectgrid = new Bucket[nb[0]*nb[1]];
  eventgrid.resize ( ( nb[0]+1 ) * ( nb[1]+1 ) );
  bl[0] = bb->get_lenght ( 0 ) / ( double ) nb[0];
  bl[1] = bb->get_lenght ( 1 ) / ( double ) nb[1];

//   // set bounds
//   double *start = bb->get_min();
//   for ( int i = 0; i < nb[0]; i++ )
//     for ( int j = 0; j < nb[1]; j++ )
//     {
//       int id = get_index ( i, j );
//       objectgrid[id].set_bbox(start[0]+i*bl[0], start[1]+j*bl[1], start[0]+(i+1)*bl[0], start[1]+(j+1)*bl[1]);
//     }

//   for ( int i = 0; i < nb[0]; i++ )
//     for ( int j = 0; j < nb[1]; j++ )
//     {
//
//     }

  tprintf ( "Init grid --> %d * %d ( dx %lf, dy %lf )\n", nb[0], nb[1], bl[0], bl[1] );
}

void Grid2d::compute_exit ( int gid, double* orig, double* dir, int &ngid, double* exit )
{
  int i, j;
  get_indices ( gid, i, j );
  double minc[2], maxc[2];
  get_cell ( i, j, minc, maxc );
  // deal with 0 cases
  if ( fabs ( dir[1] ) < Parameters::instance()->grid.epsilon ) // horizontal
  {
    if ( dir[0] > 0.0 )
    {
      exit[0] = maxc[0];
      ngid = get_index ( i+1, j );
    }
    else
    {
      exit[0] = minc[0];
      ngid = get_index ( i-1, j );
    }
    exit[1] = orig[1];
    return;
  }
  if ( fabs ( dir[0] ) < Parameters::instance()->grid.epsilon ) // vertical
  {
    exit[0] = orig[0];
    if ( dir[1] > 0.0 )
    {
      exit[1] = maxc[1];
      ngid = get_index ( i, j+1 );
    }
    else
    {
      exit[1] = minc[1];
      ngid = get_index ( i, j-1 );
    }
    return;
  }
  dbgprintf ( 5, "general case : %.20lf %.20lf\n", dir[0], dir[1] );
  // deal with general cases
  double a = dir[1]/dir[0];
  double b = orig[1]-a*orig[0];
  dbgprintf ( 5, "exit computation : %lf %lf\n", a, b );
  if ( dir[0] > 0.0 )
  {
    exit[0] = maxc[0];
    ngid = get_index ( i+1, j );
  }
  else
  {
    exit[0] = minc[0];
    ngid = get_index ( i-1, j );
  }

  exit[1] = a*exit[0] + b;

  if ( dir[1] > 0.0 && exit[1] > maxc[1] )
  {
    exit[1] = maxc[1];
    exit[0] = ( exit[1] - b ) /a;
    ngid = get_index ( i, j+1 );
  }
  else if ( dir[1] < 0.0 && exit[1] < minc[1] )
  {
    exit[1] = minc[1];
    exit[0] = ( exit[1] - b ) /a;
    ngid = get_index ( i, j-1 );
  }
}

void Grid2d::compute_crossed_bucket ( double* start, double *end, std::vector<int> &blist )
{
  int bstart = get_index ( start );
  dbgprintf ( 2, "index start %d\n", bstart );
  int bend = get_index ( end );
  dbgprintf ( 2, "index end %d\n", bend );

  double dir[2];
  geom::vec2d ( start, end, dir );
  int gid = bstart;
  double pos[2];
  int nbuc = 0; // count crossed bucket

  dbgprintf ( 2, "--> bucket" );
  while ( gid != bend )
  {
    blist.push_back ( gid );
    dbgprintf ( 2, " %d", gid );
    int ngid;
    compute_exit ( gid, start, dir, ngid, pos );
    gid = ngid;
    nbuc++;
  }
  blist.push_back ( gid ); // for the last one
}

int Bucket::potential_crash ( SweepEdge *se, SweepFacet *sf, double *crashpos )
{
  dbgprintf ( 2, "checking potential crash between se %p and sf %p\n", se, sf );
  if ( geom::line3d_plane_intersection ( se->get_start()->p, se->get_dir3d(), sf->get_plane(), Parameters::instance()->grid.epsilon, crashpos ) && geom::sqrtdistance2d ( se->get_start()->p, crashpos ) > Parameters::instance()->grid.epsilon )
  {
    dbgprintf ( 5, "--> EDGE - FACET intersection found\n" );
    if ( sf->get_frontline()->valid_future_event ( crashpos ) )
      return 1;
  }
  return 0;
}

int Bucket::potential_crash ( SweepEdge *se, SweepEdge *se1, double *crashpos )
{
  dbgprintf ( 2, "checking potential crash between se %p and se %p\n", se, se1 );
  if ( geom::lines2d_intersection ( se->get_start()->p, se1->get_start()->p, se->get_dir2d(), se1->get_dir2d(), Parameters::instance()->grid.epsilon, crashpos ) && geom::sqrtdistance2d ( se->get_start()->p, crashpos ) > Parameters::instance()->grid.epsilon )
  {
    dbgprintf ( 5, "--> EDGE - EDGE intersection found\n" );
    return 1;
  }
  return 0;
}

int Bucket::potential_crash ( SweepEdge *se, Border *bo, double *crashpos )
{
  dbgprintf ( 2, "checking potential crash between se %p and bo %p\n", se, bo );
	int singular[2] = {0,0};
  if ( geom::lines2d_intersection ( se->get_start()->p, bo->get_point ( 0 )->p, se->get_dir2d(), bo->get_dir(), Parameters::instance()->grid.epsilon, crashpos )
       && geom::sqrtdistance2d ( se->get_start()->p, crashpos ) > Parameters::instance()->grid.epsilon
       && geom::point_inside_segment2d ( bo->get_point ( 0 )->p, bo->get_point ( 1 )->p, Parameters::instance()->vorosweep.epsilon, crashpos, singular ) )
  {
    dbgprintf ( 5, "--> EDGE - BORDER intersection found\n" );
    return 1;
  }
  return 0;
}

int Bucket::potential_crash ( double *start0, double *start1, double *dir0, double *dir1, double *crashpos )
{
  if ( dir0[0]*dir1[0] < 0.0 && geom::lines2d_intersection ( start0, start1, dir0, dir1, Parameters::instance()->grid.epsilon, crashpos ) )
  {
    dbgprintf ( 5, "--> EDGE - EDGE intersection found\n" );
    return 1;
  }
  return 0;
}

void Bucket::generate_facetcrash_events ( SweepFacet *sf, double ctime, int gid, Container<Event> *econt, Event **eventlist, int &nbe )
{
// 	nbe = 0;
  if ( edgelist.empty() )
    return;

  dbgprintf ( 2, "edgelist size --> %d\n", ( int ) edgelist.size() );
  std::set<SweepEdge *>::iterator it;
  for ( it = edgelist.begin(); it != edgelist.end(); it++ )
  {
    SweepEdge *se = *it;
    double crashpos[3], crashtime;

//     if (se->is_stopped()) // TODO --> this speedup the algorithm
//     {
//       edgelist.erase(it);
//     }

    if ( !se->is_stopped() && se->is_init() && se->compatible_event ( sf ) && potential_crash ( se, sf, crashpos ) )
    {
      dbgprintf ( 2, "EDGE - FACET : crashpos --> %lf %lf %lf\n", crashpos[0], crashpos[1], crashpos[2] );
      crashtime = se->time_to_position ( crashpos );
      dbgprintf ( 2, "crashtime --> %lf\n", crashtime );
      dbgprintf ( 2, "ctime : %lf < crashtime : %lf < next switch : %lf\n", ctime, crashtime, se->get_next_switch() );
      if ( crashtime >= ctime && crashtime < se->get_next_switch() )
      {
        sf->print();
        se->print();

        Event *ev = get_matching_event ( crashpos, crashtime );
        if ( ev )
        {
          ev->append_extra ( se );
        }
        else
        {
          Event *ne = econt->get();
          ne->init ( se, sf, crashpos, crashtime, Event::EDGE_FACETCRASH, gid );
          eventlist[nbe++] = ne;
          add_event ( ne );
        }
      }
    }
  }
}

void Bucket::generate_edgecrash_events ( SweepEdge *se, double ctime, int gid, Container<Event> *econt, Event **eventlist, int &nbe )
{
  if ( !Parameters::instance()->vorosweep.multicrash )
    return;

// 	nbe = 0;
  if ( edgelist.empty() )
    return;

  dbgprintf ( 2, "edgelist size --> %d\n", ( int ) edgelist.size() );
  std::set<SweepEdge *>::iterator it;
  for ( it = edgelist.begin(); it != edgelist.end(); it++ )
  {
    SweepEdge *se1 = *it;
//     se->print();
    double crashpos[2], crashtime;

    if ( !se1->is_stopped() && se1->is_init() && potential_crash ( se, se1, crashpos ) )
    {
      dbgprintf ( 2, "EDGE - EDGE : crashpos --> %lf %lf\n", crashpos[0], crashpos[1] );
      crashtime = se->time_to_position ( crashpos );
      dbgprintf ( 2, "crashtime --> %lf\n", crashtime );
      dbgprintf ( 2, "ctime : %lf < crashtime : %lf < next switch : %lf\n", ctime, crashtime, se1->get_next_switch() );

      if ( fabs ( crashtime-se1->time_to_position ( crashpos ) ) < Parameters::instance()->vorosweep.epsilon &&
           crashtime >= ctime &&
           crashtime < se1->get_next_switch() )
      {
        se->print();
        se1->print();

        Event *ev = get_matching_event ( crashpos, crashtime );
        if ( ev )
        {
          ev->append_extra ( se );
          ev->append_extra ( se1 );
        }
        else
        {
          Event *ne = econt->get();
          ne->init ( se, se1, crashpos, crashtime, Event::EDGE_EDGECRASH2, gid );
          eventlist[nbe++] = ne;
          add_event ( ne );
        }
      }
    }
  }
}

void Bucket::generate_borderedgecrash_events ( SweepEdge *se, double ctime, int gid, Container<Event> *econt, Event **eventlist, int &nbe )
{
  // 	nbe = 0;
  if ( edgelist.empty() )
    return;

  dbgprintf ( 2, "edgelist size --> %d\n", ( int ) edgelist.size() );

  BorderSweepEdge *bse = reinterpret_cast<BorderSweepEdge *> ( se );
  Border *b0 = bse->get_border();

  double sept2d[2], sedir2d[2];

  b0->get_grid()->edge3d_to_edge2d ( bse, sept2d, sedir2d );
  std::set<SweepEdge *>::iterator it;
  for ( it = edgelist.begin(); it != edgelist.end(); it++ )
  {
    SweepEdge *se1 = *it;
//     printf ( "edge %lf %lf --> %lf %lf\n", se1->get_start()->p[0], se1->get_start()->p[1], se1->get_dir2d() [0], se1->get_dir2d() [1] );
    if ( !se1->is_stopped() && se1->get_type() == SweepObject::BORDEREDGE )
    {
      dbgprintf ( 2, "--> found 1 border edge %p\n", se1 );
      BorderSweepEdge *bse1 = reinterpret_cast<BorderSweepEdge *> ( se1 );
      Border *b1 = bse1->get_border();

      if ( b0 == b1 )
      {
        double crashpos2d[3], crashtime;
        double s2pt2d[2], s2dir2d[2];

        b1->get_grid()->edge3d_to_edge2d ( bse1, s2pt2d, s2dir2d );
        if ( potential_crash ( sept2d, s2pt2d, sedir2d, s2dir2d, crashpos2d ) )
        {
          double crashpos3d[3];
          b1->get_grid()->pt2d_to_pt3d ( crashpos2d, crashpos3d );
          dbgprintf ( 2, "EDGE - EDGE : crashpos --> %lf %lf %lf\n", crashpos3d[0], crashpos3d[1], crashpos3d[2] );
          crashtime = bse->time_to_position ( crashpos3d );
          dbgprintf ( 2, "crashtime --> %lf\n", crashtime );
          dbgprintf ( 2, "ctime : %lf < crashtime : %lf < next switch : %lf\n", ctime, crashtime, se1->get_next_switch() );
          if ( fabs ( crashtime-se1->time_to_position ( crashpos3d ) ) < Parameters::instance()->vorosweep.epsilon &&
               crashtime >= ctime &&
               crashtime < se1->get_next_switch() )
          {
            Event *ev = get_matching_event ( crashpos3d, crashtime );
            if ( ev )
            {
              if ( se->get_xyspeed() == INFINITY )
              {
                double dist = geom::sqrtdistance2d ( se->get_start()->p, crashpos3d );
                se->update_event ( ev, ctime, dist );
              }
              else if ( se1->get_xyspeed() == INFINITY )
              {
                double dist = geom::sqrtdistance2d ( se1->get_start()->p, crashpos3d );
                se1->update_event ( ev, ctime, dist );
              }
              ev->append_extra ( se );
              ev->append_extra ( se1 );
            }
            else
            {
              Event *ne = econt->get();
              ne->init ( se, se1, crashpos3d, crashtime, Event::EDGE_EDGECRASH1, gid );
              eventlist[nbe++] = ne;
              add_event ( ne );
            }
          }
        }
      }
    }
  }
}

// void Bucket::generate_bordercornercrash_events ( SweepEdge *se, double ctime, int gid, Container<Event> *econt, Event **eventlist, int &nbe )
// {
//   // 	nbe = 0;
//   dbgprintf ( 2, "borderlist size --> %d\n", ( int ) borderlist.size() );
//
//   if ( se->is_stopped() || borderlist.empty() )
//     return;
//
//   BorderSweepEdge *bse = reinterpret_cast<BorderSweepEdge *> ( se );
//   Border *bo = bse->get_border();
//
//   Border b1 = NULL;
//   if ( geom::cross_product2d ( se->get_dir2d(), bo->get_dir() ) >= 0.0 )
//     b1 = bo->get_adj ( 0 );
//   else
//     b1 = bo->get_adj ( 1 );
//
//   double crashpos[2], crashtime;
//
//   potential_crash ( se, bo, crashpos );
//   dbgprintf ( 2, "EDGE - BORDER : crashpos --> %lf %lf\n", crashpos[0], crashpos[1] );
//   crashtime = se->time_to_position ( crashpos );
//   dbgprintf ( 2, "crashtime --> %lf\n", crashtime );
//   if ( crashtime < se->get_next_switch() )
//   {
//     Event *ne = econt->get();
//     ne->init ( se, bo, crashpos, crashtime, Event::EDGE_BORDERCRASH, gid );
//     eventlist[nbe++] = ne;
//   }
// }

void Bucket::generate_facetcrash_events ( SweepEdge *se, double ctime, int gid, Container<Event> *econt, Event **eventlist, int &nbe )
{
// 	nbe = 0;

  dbgprintf ( 2, "facetlist size --> %d\n", ( int ) facetlist.size() );
  if ( se->is_stopped() || !se->is_init() || facetlist.empty() )
    return;

  std::set<SweepFacet *>::iterator it;
  for ( it = facetlist.begin(); it != facetlist.end(); it++ )
  {
    SweepFacet *sf = ( *it );
//     sf->print();
    double crashpos[3], crashtime;

    if ( se->compatible_event ( sf ) && potential_crash ( se, sf, crashpos ) )
    {
      dbgprintf ( 2, "EDGE - FACET : crashpos --> %lf %lf %lf\n", crashpos[0], crashpos[1], crashpos[2] );
      crashtime = se->time_to_position ( crashpos );
      dbgprintf ( 2, "crashtime --> %lf\n", crashtime );
      dbgprintf ( 2, "ctime : %lf < crashtime : %lf < next switch : %lf\n", ctime, crashtime, se->get_next_switch() );
      if ( crashtime >= ctime && crashtime < se->get_next_switch() )
      {
        sf->print();
        se->print();

        Event *ev = get_matching_event ( crashpos, crashtime );
        if ( ev )
        {
          if ( se->get_xyspeed() == INFINITY )
          {
            double dist = geom::sqrtdistance2d ( se->get_start()->p, crashpos );
            se->update_event ( ev, ctime, dist );
          }
          ev->append_extra ( se );
        }
        else
        {
          Event *ne = econt->get();
          ne->init ( se, sf, crashpos, crashtime, Event::EDGE_FACETCRASH, gid );
          eventlist[nbe++] = ne;
          add_event ( ne );
        }
      }
    }
  }
}

void Bucket::generate_bordercrash_events ( SweepEdge *se, double ctime, int gid, Container<Event> *econt, Event **eventlist, int &nbe )
{
// 	nbe = 0;
  dbgprintf ( 2, "borderlist size --> %d\n", ( int ) borderlist.size() );

  if ( se->is_stopped() || !se->is_init() || borderlist.empty() )
    return;

  std::set<Border *>::iterator it;
  for ( it = borderlist.begin(); it != borderlist.end(); it++ )
  {
    Border *bo = ( *it );
    double crashpos[2], crashtime;
    bool good_face_of_border = ( geom::cross_product2d ( se->get_dir2d(), bo->get_dir() ) >= 0.0 );
    if ( good_face_of_border && potential_crash ( se, bo, crashpos ) )
    {
      dbgprintf ( 2, "EDGE - BORDER : crashpos --> %lf %lf\n", crashpos[0], crashpos[1] );
      crashtime = se->time_to_position ( crashpos );
      dbgprintf ( 2, "crashtime --> %lf\n", crashtime );
      if ( crashtime >= ctime && crashtime < se->get_next_switch() )
      {
        Event *ne = econt->get();
        ne->init ( se, bo, crashpos, crashtime, Event::EDGE_BORDERCRASH, gid );
        eventlist[nbe++] = ne;
      }
    }
  }
}

void Bucket::generate_bordercrash_events ( SweepFacet *sf, double ctime, int gid, Container<Event> *econt, Event **eventlist, int &nbe )
{
// 	nbe = 0;
  dbgprintf ( 2, "borderpointlist size --> %d\n", ( int ) borderpointlist.size() );

  if ( borderpointlist.empty() )
    return;

  std::set<std::pair<Border *, int> >::iterator it;
//   std::set<Border *>::iterator it;
  for ( it = borderpointlist.begin(); it != borderpointlist.end(); it++ )
  {
    Border *bo = ( *it ).first;
    int idpt = ( *it ).second;

    Point *pt = bo->get_point ( idpt );
    double crashpos[3], crashtime;

    if ( vertical_crash ( pt->p, 1.0, sf, crashpos ) )
    {
      bool good_face_of_border = ( geom::cross_product2d ( sf->get_plane(), bo->get_dir() ) >= 0.0 );
      if ( good_face_of_border && sf->get_frontline()->valid_future_event ( pt->p ) )
      {
        dbgprintf ( 2, "FACET - BORDER : crashpos --> %lf %lf %lf\n", crashpos[0], crashpos[1], crashpos[2] );
        crashtime = sf->time_to_position ( crashpos );
        dbgprintf ( 2, "crashtime --> %lf\n", crashtime );
        if ( crashtime >= ctime )
        {
          if ( ( idpt == 0 && geom::scalar_product2d ( sf->get_plane(), bo->get_dir() ) > 0.0 ) ||
               ( idpt == 1 && geom::scalar_product2d ( sf->get_plane(), bo->get_dir() ) < 0.0 ) )
          {
            Event *ne = econt->get();
// 						printf("new event from scratch --> %p\n", ne);
            ne->init ( sf, bo, crashpos, crashtime, Event::FACET_BORDERCRASH, gid );
            eventlist[nbe++] = ne;
// 						Event *ev = get_matching_event ( crashpos, crashtime );
//             if ( ev )
//             {
// //               ev->append_extra ( se );
//             }
//             else
//             {
//               Event *ne = econt->get();
//               ne->init ( sf, bo, crashpos, crashtime, Event::FACET_BORDERCRASH, gid );
//               eventlist[nbe++] = ne;
//               add_event ( ne );
//             }
          }
        }
      }
    }
  }
}

int Bucket::vertical_crash ( double *pt, double z, SweepFacet *sf, double *crashpos )
{
  static double dir[3] = {0.0, 0.0, 1.0}; // vertical ray
  dir[2] = z;
  if ( geom::line3d_plane_intersection ( pt, dir, sf->get_plane(), Parameters::instance()->grid.epsilon, crashpos ) )
  {
//     printf ( "crashpos --> %lf %lf %lf\n", crashpos[0], crashpos[1], crashpos[2] );
    if ( geom::sqrtdistance2d ( sf->get_start()->p, crashpos ) > Parameters::instance()->grid.epsilon )
    {
      dbgprintf ( 2, "--> VERTICAL RAY - FACET intersection found\n" );
      return 1;
    }
  }
  return 0;
}

int Bucket::point_already_covered ( double *pt, double ctime )
{
  if ( facetlist.empty() )
    return 0;
  dbgprintf ( 2, "facetlist size --> %d\n", ( int ) facetlist.size() );
  std::set<SweepFacet *>::iterator it;
  for ( it = facetlist.begin(); it != facetlist.end(); it++ )
  {
    SweepFacet *sf = ( *it );
    double crashpos[3];
    if ( vertical_crash ( pt, -1.0, sf, crashpos ) )
    {
      dbgprintf ( 2, "VERTICAL RAY - FACET : crashpos --> %lf %lf %lf\n", crashpos[0], crashpos[1], crashpos[2] );
      if ( crashpos[2] > 0.0 && crashpos[2] < ctime )
      {
        return 1;
      }
    }
  }
  return 0;
}

void Grid2d::export_vtk ( const char *name )
{
  char filename[100];
  static int count = 0;
  sprintf ( filename, "%s_%d.vtk", name, count++ );

  tprintf ( "Writing file --> %s\n", filename );
  FILE *fp_rw = fopen ( filename, "w" );

  fprintf ( fp_rw, "# vtk DataFile Version 2.0\n" );
  fprintf ( fp_rw, "Really cool data\n" );
  fprintf ( fp_rw, "ASCII\n" );
  fprintf ( fp_rw, "DATASET POLYDATA\n" );
  fprintf ( fp_rw, "POINTS %d float\n", ( nb[0]+1 ) * ( nb[1]+1 ) );

  for ( int j = 0; j <= nb[1]; j++ )
    for ( int i = 0; i <= nb[0]; i++ )
    {
      fprintf ( fp_rw, "%lf %lf %lf\n", bb->get_min() [0] + i*bl[0], bb->get_min() [1] + j*bl[1], 0.0 );
    }

  fprintf ( fp_rw, "POLYGONS %d %d\n", nb[0]*nb[1], 5*nb[0]*nb[1] );
  for ( int j = 0; j < nb[1]; j++ )
    for ( int i = 0; i < nb[0]; i++ )
    {
      fprintf ( fp_rw, "%d %d %d %d %d\n", 4, i+j* ( nb[0]+1 ), i+1+j* ( nb[0]+1 ), i+1+ ( j+1 ) * ( nb[0]+1 ), i+ ( j+1 ) * ( nb[0]+1 ) );
    }

  fprintf ( fp_rw, "CELL_DATA %d\n", nb[0]*nb[1] );
  fprintf ( fp_rw, "SCALARS index int 1\n" );
  fprintf ( fp_rw, "LOOKUP_TABLE default\n" );
  for ( int j = 0; j < nb[1]; j++ )
    for ( int i = 0; i < nb[0]; i++ )
    {
      fprintf ( fp_rw, "%d\n", get_index ( i, j ) );
    }

  fprintf ( fp_rw, "SCALARS nb_faces int 1\n" );
  fprintf ( fp_rw, "LOOKUP_TABLE default\n" );
  for ( int j = 0; j < nb[1]; j++ )
    for ( int i = 0; i < nb[0]; i++ )
    {
      fprintf ( fp_rw, "%d\n", get_bucket ( get_index ( i, j ) )->get_sweepfacet_size() );
    }

  fclose ( fp_rw );
}


