// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"
#include "border.h"

using namespace vorosweep;

void Graph::handle_borderedgeswitch ( Event *ev )
{
  SweepEdge *se = ev->get_sweepedge();

  if ( se->is_stopped() )
    return;

  if ( check_T_configuration ( se ) )
    return;

  double *pos = ev->get_coord();
  double time = ev->get_time();

  se->update_current ( time, pos );

  int gid = ev->get_gid();
  se->set_last_gid ( gid );

  Border *bo = ev->get_border();
  bo->print();
  {
    if ( check_for_corner ( se, bo, time ) )
      return;

    switch_border ( se, bo, time );
    return;
  }
}

int Graph::check_for_corner ( SweepEdge *se, Border *bo, double time )
{
  if ( !se->get_front ( 0 ) )
  {
    SweepEdge *se1 = se->get_front ( 1 )->get_sweepedge ( 1 );
    if ( se1->get_type() == SweepObject::BORDEREDGE )
    {
      Border *cbo = reinterpret_cast<BorderSweepEdge *> ( se1 )->get_border();
      if ( cbo == bo->get_adj ( 1 ) )
      {
        se->set_stopped();
        se1->set_stopped();
        return 1;
      }
    }
  }
  else if ( !se->get_front ( 1 ) )
  {
    SweepEdge *se0 = se->get_front ( 0 )->get_sweepedge ( 0 );
    if ( se0->get_type() == SweepObject::BORDEREDGE )
    {
      Border *cbo = reinterpret_cast<BorderSweepEdge *> ( se0 )->get_border();
      if ( cbo == bo->get_adj ( 0 ) )
      {
        se->set_stopped();
        se0->set_stopped();
        return 1;
      }
    }
  }
  return 0;
}

void Graph::switch_border ( SweepEdge *se, Border *bo, double time )
{
  // find the direction of the next borderbucket
  dbgprintf ( 2, "--> switching border\n" );
  Border *nextborder = NULL;
  SweepFacet *sf = NULL;
  int bgid = se->get_last_gid();
  BorderSweepEdge *nse = NULL;
  double ndir[3];
  if ( geom::scalar_product2d ( se->get_dir2d(), bo->get_dir() ) > 0.0 )
  {
		if (bo->has_point_been_reached ( 1 ))
			return;
		bo->set_point_reached ( 1 );
		
    nextborder = bo->get_adj ( 0 );		
    sf = se->get_front ( 0 )->get_owner();
		
    sf->dir2d_to_dir3d ( nextborder->get_dir(), ndir );
    double *sedir = sf->get_angular_bound ( 1 );
    predicate::is_horizontal ( ndir );
    dbgprintf ( 2, "cross --> %lf\n", geom::cross_product2d ( sedir, ndir ) );
    // simplest case
    if ( ndir[2] >= 0.0 && geom::cross_product2d ( sedir, ndir ) >= 0.0 )
    {
      nse = new BorderSweepEdge ( se->get_current(), ndir, time );
    }
    else
    {
      singular_point_rotation_cw ( se, nextborder, time );
      return;
    }
  }
  else
  {
	  if (bo->has_point_been_reached ( 0 ))
			return;
		bo->set_point_reached ( 0 );
		
    nextborder = bo->get_adj ( 1 );
    sf = se->get_front ( 1 )->get_owner();
//     bgid = nextborder->get_last_bucket_index();
		bgid = se->get_last_gid();
    sf->dir2d_to_dir3d ( nextborder->get_dir(), ndir );
    double *sedir = sf->get_angular_bound ( 0 );
    geom::multvec3d ( ndir, -1.0 );
    predicate::is_horizontal ( ndir );
    dbgprintf ( 2, "cross --> %lf\n", geom::cross_product2d ( ndir, sedir ) );
    // simplest case -->
    if ( ndir[2] >= 0.0 && geom::cross_product2d ( ndir, sedir ) >= 0.0 )
    {
      nse = new BorderSweepEdge ( se->get_current(), ndir, time );
    }
    else
    {
      singular_point_rotation_ccw ( se, nextborder, time );
      return;
    }
  }

  // stopping se
  se->set_stopped();

  if ( nse )
  {
    nse->set_border ( nextborder );
    nse->update_speeds ( sf->get_plane() );
    FrontLine *fl = sf->get_frontline();
    fl->update_sweepedge ( se, nse );
    Event *ne = econtainer->get();
    ne->init ( nse, nextborder, nse->get_start()->p, time, Event::NEWEDGESWITCH, bgid );
    add_priority_event ( ne );
  }
}

// this routine generate as much as facet as needed to circumvent obstacle
void Graph::singular_point_rotation_ccw ( SweepEdge *se, Border *nextborder, double time )
{
  dbgprintf ( 2, "--> COUNTERCLOCKWISE rotation\n" );
  SweepFacet *sf = se->get_front ( 1 )->get_owner();
  Generator *g = sf->get_generator();

  int gid = bucketgrid->get_index ( se->get_current()->p );

  double ndir[3];

  int index = sf->get_pos();
  int nbfacet = 0;

  double *sedir = sf->get_angular_bound ( 0 );
  SweepEdge *lastse = new SweepEdge ( se->get_current(), sedir, time );
  lastse->update_speeds ( g->get_next_normal ( index ) );
  lastse->set_init();

  // stopping se
  se->set_stopped();
  se->get_front ( 1 )->get_frontline()->update_sweepedge ( se, lastse );

  // add switch event
  Event *nes = econtainer->get();
  nes->init ( lastse, lastse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
  add_priority_event ( nes );

  double *se0dir = NULL;
  double *se1dir = NULL;

  dbgprintf ( 2, "init normal --> %lf %lf %lf\n", sf->get_plane() [0], sf->get_plane() [1], sf->get_plane() [2] );
  do
  {
    int nextindex = g->get_next_index ( index );
    double *nextnormal = g->get_normal ( nextindex );
    double *nextdir = g->get_normal ( nextindex );
    double nextspeed = g->get_speed ( nextindex );


    dbgprintf ( 2, "checking with normal --> %lf %lf %lf\n", nextnormal[0], nextnormal[1], nextnormal[2] );
//     nextsf->dir2d_to_dir3d ( nextborder->get_dir(), ndir );
    geom::dir2d_to_dir3d ( nextborder->get_dir(), nextnormal, ndir );
    geom::multvec3d ( ndir, -1.0 );
    predicate::is_horizontal ( ndir );

    SweepFacet *nsf = new SweepFacet ( se->get_current(), nextnormal, nextdir, time );
    SweepEdge *nse;

    se0dir = g->get_edge ( nextindex, 0 );
    se1dir = g->get_edge ( nextindex, 1 );

    dbgprintf ( 2, "border dir --> %lf %lf\n", nextborder->get_dir() [0], nextborder->get_dir() [1] );
    dbgprintf ( 2, "dir 0 --> %lf %lf\n", se0dir[0], se0dir[1] );
    dbgprintf ( 2, "dir 1 --> %lf %lf\n", se1dir[0], se1dir[1] );
    dbgprintf ( 2, "cross --> %lf %lf\n", geom::cross_product2d ( nextborder->get_dir(), se0dir ), geom::cross_product2d ( se1dir, nextborder->get_dir() ) );

    if ( geom::cross_product2d ( nextborder->get_dir(), se0dir ) > 0.0 || geom::cross_product2d ( se1dir, nextborder->get_dir() ) > 0.0 )
    {
      nse = new SweepEdge ( se->get_current(), se0dir, time );
      nse->set_init();
      // add switch event
      Event *nes = econtainer->get();
      nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
      add_priority_event ( nes );
    }
    else
    {
      BorderSweepEdge *nbse = new BorderSweepEdge ( se->get_current(), ndir, time );
      nbse->set_border ( nextborder );
      nse = nbse;
      // add switch event
      Event *nes = econtainer->get();
//       nes->init ( nse, nextborder, nse->get_start()->p, time, Event::NEWBORDEREDGESWITCH, nextborder->get_last_bucket_index() );
      nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
      add_priority_event ( nes );
    }

    nsf->set_pos ( nextindex );
    nsf->set_xyspeed ( nextspeed );
    nsf->set_generator ( g );
    g->add_sweepfacet ( nsf );
    new FrontLine ( nsf, nse, lastse );
    nse->update_speeds ( nsf->get_plane() );

    // add switch event
    Event *nfs = econtainer->get();
    nfs->init ( nsf, nsf->get_start()->p, time, Event::NEWFACETSWITCH, gid );
    add_priority_event ( nfs );

    index = nextindex;
    lastse = nse;
		
		ASSERT_ERROR ( nbfacet <= g->initial_face_size(), "full rotation without stopping at %lf %lf\n", nse->get_current()->p[0], nse->get_current()->p[1] );
		
    nbfacet++;
  }
  while ( geom::cross_product2d ( nextborder->get_dir(), se0dir ) > 0.0 || geom::cross_product2d ( se1dir, nextborder->get_dir() ) > 0.0 );

}

// this routine generate as much as facet as needed to circumvent obstacle
void Graph::singular_point_rotation_cw ( SweepEdge *se, Border *nextborder, double time )
{
  dbgprintf ( 2, "--> CLOCKWISE rotation\n" );
  SweepFacet *sf = se->get_front ( 0 )->get_owner();
  Generator *g = sf->get_generator();

  int gid = bucketgrid->get_index ( se->get_current()->p );

  double ndir[3];

  int index = sf->get_pos();
  int nbfacet = 0;
  double *sedir = sf->get_angular_bound ( 1 );
  SweepEdge *lastse = new SweepEdge ( se->get_current(), sedir, time );
  lastse->update_speeds ( g->get_prev_normal ( index ) );
  lastse->set_init();

  // stopping se
  se->set_stopped();
  se->get_front ( 0 )->get_frontline()->update_sweepedge ( se, lastse );

  // add switch event
  Event *nes = econtainer->get();
  nes->init ( lastse, lastse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
  add_priority_event ( nes );

//   SweepEdge *se0 = NULL;
//   SweepEdge *se1 = NULL;

  double *se0dir = NULL;
  double *se1dir = NULL;

  dbgprintf ( 2, "init normal --> %lf %lf %lf\n", sf->get_plane() [0], sf->get_plane() [1], sf->get_plane() [2] );
  do
  {
    int previndex = g->get_prev_index ( index );
    double *prevnormal = g->get_normal ( previndex );
    double *prevdir = g->get_dir ( previndex );
    double prevspeed = g->get_speed ( previndex );

    dbgprintf ( 2, "checking with normal --> %lf %lf %lf\n", prevnormal[0], prevnormal[1], prevnormal[2] );
    geom::dir2d_to_dir3d ( nextborder->get_dir(), prevnormal, ndir );
    dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );

    SweepFacet *nsf = new SweepFacet ( se->get_current(), prevnormal, prevdir, time );
    SweepEdge *nse;

    se0dir = g->get_edge ( previndex, 0 );
    se1dir = g->get_edge ( previndex, 1 );

    dbgprintf ( 2, "border dir --> %lf %lf\n", nextborder->get_dir() [0], nextborder->get_dir() [1] );
    dbgprintf ( 2, "dir 0 --> %lf %lf\n", se0dir[0], se0dir[1] );
    dbgprintf ( 2, "dir 1 --> %lf %lf\n", se1dir[0], se1dir[1] );
    dbgprintf ( 2, "cross --> %lf %lf\n", geom::cross_product2d ( nextborder->get_dir(), se0dir ), geom::cross_product2d ( se1dir, nextborder->get_dir() ) );

    if ( geom::cross_product2d ( nextborder->get_dir(), se0dir ) < 0.0 || geom::cross_product2d ( se1dir, nextborder->get_dir() ) < 0.0 )
    {
      nse = new SweepEdge ( se->get_current(), se1dir, time );
      nse->set_init();
      // add switch event
      Event *nes = econtainer->get();
      nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
      add_priority_event ( nes );
    }
    else
    {
      BorderSweepEdge *nbse = new BorderSweepEdge ( se->get_current(), ndir, time );
      nbse->set_border ( nextborder );
      nse = nbse;
      // add switch event
      Event *nes = econtainer->get();
//       nes->init ( nse, nextborder, nse->get_start()->p, time, Event::NEWBORDEREDGESWITCH, nextborder->get_first_bucket_index() );
      nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
      add_priority_event ( nes );
    }

    nsf->set_pos ( previndex );
    nsf->set_xyspeed ( prevspeed );
    nsf->set_generator ( g );
    g->add_sweepfacet ( nsf );
    new FrontLine ( nsf, lastse, nse );
    nse->update_speeds ( nsf->get_plane() );

    // add switch event
    Event *nfs = econtainer->get();
    nfs->init ( nsf, nsf->get_start()->p, time, Event::NEWFACETSWITCH, gid );
    add_priority_event ( nfs );

    index = previndex;
    lastse = nse;
		
		ASSERT_ERROR ( nbfacet <= g->initial_face_size(), "full rotation without stopping at %lf %lf\n", nse->get_current()->p[0], nse->get_current()->p[1] );
		
    nbfacet++;
  }
  while ( geom::cross_product2d ( nextborder->get_dir(), se0dir ) < 0.0 || geom::cross_product2d ( se1dir, nextborder->get_dir() ) < 0.0 );

}

