// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"
#include "border.h"

using namespace vorosweep;

void Graph::handle_edge_bordercrash ( Event *ev )
{
  SweepEdge *se = ev->get_sweepedge();

  if ( se->is_stopped() )
    return;

  Border *bo = ev->get_border();
  bo->print();

//   Grid1d *bordergrid = bo->get_grid();	
  dbgprintf ( 5, "--> border crash\n" );
  double time = ev->get_time();
  double *pos = ev->get_coord();
  int gid = ev->get_gid();

  if ( se->is_init() ) // case sweepedge is init : generate 1 or 2 edges
  {
    se->update_current ( time, pos );
    se->set_stopped();

    for ( int i = 0; i < 2; i++ )
    {
      SweepFacet *sf = se->get_sweepfacet ( i );
//       printf ( "sf --> %p\n", sf );
      if ( sf )
      {
        sf->print();
        // compute new directions
        double ndir[3];
        if ( i == 0 )
          new_dir ( bo->get_plane(), sf->get_plane(), ndir );
        else
          new_dir ( sf->get_plane(), bo->get_plane(), ndir );

        geom::normalize3d ( ndir );
				predicate::is_horizontal ( ndir );	

        if ( ndir[2] >= 0.0 )
        {
          BorderSweepEdge *nse = new BorderSweepEdge ( se->get_current(), ndir, time );
          nse->set_border ( bo );
          nse->update_speeds ( sf->get_plane() );
          FrontLine *fl = sf->get_frontline();
          fl->update_sweepedge ( se, nse );

          Event *ne = econtainer->get();
          ne->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
          add_priority_event ( ne );
        }
        else
        {
          dbgprintf ( 2, "--> closing border front \n" );
          Front *fr = se->get_front ( i );
          if ( fr )
          {
            SweepEdge *secomp = fr->get_sweepedge ( i );
            se->merge_current ( secomp, get_time() );
            // close the front
            FrontLine *cfl = fr->get_frontline();
            cfl->close_front ( fr );
            secomp->set_stopped();
          }
        }
      }
    }
  }
  else // merge other edges
  {
    for ( int i = 0; i < 2; i++ )
    {
      Front *fr = se->get_front ( i );
      if ( fr )
      {
        SweepEdge *secomp = fr->get_sweepedge ( i );
        se->merge_current ( secomp, get_time() );
        // close the front
        FrontLine *cfl = fr->get_frontline();
        cfl->close_front ( fr );
      }
    }
  }
  se->set_stopped();
}
