// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"

using namespace vorosweep;

void Graph::handle_edge_edgecrash1 ( Event *ev )
{
  SweepEdge *se0 = ev->get_sweepedge ( 0 );
  SweepEdge *se1 = ev->get_sweepedge ( 1 );

  se0->print();
  se1->print();
	
	double *pos = ev->get_coord();
  double time = ev->get_time();

  // if one of the edges is still running
  // we have to add its earliest event in the main queue
  if ( se0->is_stopped() && !se1->is_stopped() )
  {
    if ( !check_T_configuration ( se1 ) )
      add_earliest_event ( se1 );
    return;
  }
  else if ( !se0->is_stopped() && se1->is_stopped() )
  {
    if ( !check_T_configuration ( se0 ) )
      add_earliest_event ( se0 );
    return;
  }
  else if ( se0->is_stopped() && se1->is_stopped() )
  {
    return;
  }
  
  update_time ( time );

  se0->update_current ( time, pos );
  se1->update_current ( time, pos );

  if ( se0->get_xyspeed() <= se1->get_xyspeed() ) // take speed into account
    se0->merge_current ( se1, time ); // stop and update as well (consistency)
  else
    se1->merge_current ( se0, time );

  if ( !se0->get_front ( 0 ) && !se1->get_front ( 1 ) )
  {
    merge ( se1, se0 );
  }
  else if ( !se0->get_front ( 1 ) && !se1->get_front ( 0 ) )
  {
    merge ( se0, se1 );
  }
  else if ( se0->get_front ( 0 ) == se1->get_front ( 1 ) )
  {
    merge ( se1, se0 );
  }
  else
  {
    merge ( se0, se1 );
  }
  
  se0->print();
	se1->print();
}

int Graph::check_T_configuration ( SweepEdge *se )
{
  Front *fr0 = se->get_front ( 0 );
  Front *fr1 = se->get_front ( 1 );

  if ( fr0 && fr1 )
  {
    if ( fr0->get_sweepedge ( 0 )->is_stopped() && fr1->get_sweepedge ( 1 )->is_stopped() ) // "T" configuration
    {
			SweepEdge *secomp = fr0->get_sweepedge ( 0 );
			secomp->merge_current(se);
      return 1;
    }
  }
  else if ( fr0 && !fr1 && fr0->get_sweepedge ( 0 )->is_stopped() )
  {
    SweepEdge *secomp = fr0->get_sweepedge ( 0 );
    // since secomp is already stopped, secomp->current has already been merged so is referenced by several edges and
    // we cannot remove secomp->current. Instead, we merge se in secomp
    secomp->merge_current ( se );
    return 1;
  }
  if ( !fr0 && fr1 && fr1->get_sweepedge ( 1 )->is_stopped() )
  {
    SweepEdge *secomp = fr1->get_sweepedge ( 1 );
    secomp->merge_current ( se );
    return 1;
  }
  return 0;
}

