// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"

using namespace vorosweep;

void Graph::handle_edge_edgecrash2 ( Event *ev )
{
  SweepEdge *se0 = ev->get_sweepedge ( 0 );
  SweepEdge *se1 = ev->get_sweepedge ( 1 );

  se0->print();
  se1->print();

  // if one of the edges is still running
  // we have to add its earliest event in the main queue
  if ( se0->is_stopped() && !se1->is_stopped() )
  {
    Front *fr0 = se1->get_front ( 0 );
    Front *fr1 = se1->get_front ( 1 );
    if ( fr0 && fr1 )
    {
      if ( fr0->get_sweepedge ( 0 )->is_stopped() && fr1->get_sweepedge ( 1 )->is_stopped() ) // "T" configuration
      {
        se1->set_stopped();
        return;
      }
    }
    add_earliest_event ( se1 );
    return;
  }
  else if ( !se0->is_stopped() && se1->is_stopped() )
  {
    Front *fr0 = se0->get_front ( 0 );
    Front *fr1 = se0->get_front ( 1 );

    if ( fr0 && fr1 )
    {
      if ( fr0->get_sweepedge ( 0 )->is_stopped() && fr1->get_sweepedge ( 1 )->is_stopped() ) // "T" configuration
      {
        se0->set_stopped();
        return;
      }
    }
    add_earliest_event ( se0 );
    return;
  }
  else if ( se0->is_stopped() && se1->is_stopped() )
  {
    return;
  }

  double *pos = ev->get_coord();
  double time = ev->get_time();
  update_time ( time );

  int gid = bucketgrid->get_index ( pos );

  SweepFacet *sf[2][2];

  sf[0][0] = se0->get_sweepfacet ( 0 );
  sf[0][1] = se1->get_sweepfacet ( 1 );
  sf[1][0] = se1->get_sweepfacet ( 0 );
  sf[1][1] = se0->get_sweepfacet ( 1 );

  double ndir[2][3];

  for ( int i = 0; i < 2; i++ )
  {
		new_dir(sf[i][0]->get_plane(), sf[i][1]->get_plane(), ndir[i]);
    geom::normalize3d ( ndir[i] );
    // check new dir
    if ( ndir[i][2] < 0.0 )
    {
      dbgprintf ( 2, "ndir[2] < 0.0 --> return\n" );
      return;
    }
  }
  
  if ( se0->get_xyspeed() <= se1->get_xyspeed() ) // take speed into account
    se0->merge_current ( se1, time ); // stop and update as well (consistency)
  else
    se1->merge_current ( se0, time );
	
// 	se0->merge_current ( se1, time );

// create new edges
  for ( int i = 0; i < 2; i++ )
  {
//     SweepFacet *sf = se->get_sweepfacet ( i );
    if ( sf[i][0] )
    {
// 			predicate::is_horizontal ( ndir[i] );
//       if ( ndir[i][2] >= Parameters::instance()->vorosweep.epsilon || fabs ( ndir[i][2] ) < Parameters::instance()->vorosweep.epsilon )
//       if ( ndir[i][2] >= -Parameters::instance()->sweepobject.horizontal )
			if ( ndir[i][2] >= 0.0 )
      {
//         if ( fabs ( ndir[i][2] ) < Parameters::instance()->sweepobject.horizontal )
//         {
//           dbgprintf ( 2, "--> horizontal dir\n" );
//           ndir[i][2] = 0.0;
//         }
//         predicate::is_horizontal ( ndir[i] );

        SweepEdge *nse = new SweepEdge ( se0->get_current(), ndir[i], time );

        nse->update_speeds ( sf[i][0]->get_plane() );
        if ( i ==0 )
        {
          FrontLine *fl0 = sf[i][0]->get_frontline();
          fl0->update_sweepedge ( se0, nse );
          FrontLine *fl1 = sf[i][1]->get_frontline();
          fl1->update_sweepedge ( se1, nse );
        }
        else
        {
          FrontLine *fl0 = sf[i][0]->get_frontline();
          fl0->update_sweepedge ( se1, nse );
          FrontLine *fl1 = sf[i][1]->get_frontline();
          fl1->update_sweepedge ( se0, nse );
        }

        // add switch event
        Event *nes = econtainer->get();
        nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
        add_priority_event ( nes );
      }
    }
  }
//   getchar();
}


