// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"

using namespace vorosweep;

void Graph::handle_edge_facetcrash ( Event *ev )
{
  SweepEdge *se = ev->get_sweepedge ( 0 );
  SweepFacet *msf = ev->get_sweepfacet ( 1 );
	
	if ( !msf->is_stopped() )
		add_earliest_event ( msf );
		
	if ( !se->is_stopped() )
		add_earliest_event ( se );
	
	if ( se->is_stopped() )
    return;
	
	if ( !se->is_init() )
    return;

  se->print();
  msf->print();
	
  FrontLine *mfl = msf->get_frontline();
	
  double *pos = ev->get_coord();
  double time = ev->get_time();
	int gid = ev->get_gid();

  int singular[2] = {0, 0};

  Front *mfr = mfl->get_including_front ( pos, time, singular );

  if ( !mfr ) // check if point is not included into a front of frontline
    return;

  dbgprintf ( 5, "sweepedge : %p; sweepfacet %p\n", se, msf );
  dbgprintf ( 5, "sweepedge dir --> %lf %lf\n",  se->get_dir2d() [0], se->get_dir2d() [1] );
  dbgprintf ( 5, "sweepfacet dir --> %lf %lf\n",  msf->get_dir2d() [0], msf->get_dir2d() [1] );

  dbgprintf ( 2, "sweepedge %p --> is init : %d\n",  se, se->is_init() );

	if (already_merged(se->get_front(0), mfr))
		return;
	if (already_merged(mfr, se->get_front(1)))
		return;

  // update sweepedges
  SweepEdge *sefr[2];
  sefr[0] = mfr->get_sweepedge ( 0 );
  sefr[1] = mfr->get_sweepedge ( 1 );

//   if ( sefr[0]->get_sweepfacet ( 0 ) == se->get_sweepfacet ( 0 ) || sefr[0]->get_sweepfacet ( 0 ) == se->get_sweepfacet ( 1 ) )
//     return;
//   if ( sefr[1]->get_sweepfacet ( 1 ) == se->get_sweepfacet ( 0 ) || sefr[1]->get_sweepfacet ( 1	) == se->get_sweepfacet ( 1 ) )
//     return;

  se->update_current ( time, pos );
  sefr[0]->update_current ( time );
  sefr[1]->update_current ( time );

  SweepFacet *esf[2];
  // cas general
  esf[0] = msf;
  esf[1] = msf;

  if ( singular[0])
  {
//     dbgprintf ( 1, "--> edge_facetcrash singular case 0\n" );
//     printf ( "--> edge_facetcrash singular case 0\n" );
    esf[0] = sefr[0]->get_front ( 0 )->get_owner();
  }
  else if ( singular[1])
  {
    dbgprintf ( 1,  "--> edge_facetcrash singular case 1\n" );
//     printf ( "--> edge_facetcrash singular case 1\n" );
    esf[1] = sefr[1]->get_front ( 1 )->get_owner();
  }

  double ndir[2][3];

  for ( int i = 0; i < 2; i++ )
  {
    SweepFacet *sf = se->get_sweepfacet ( i );
    if ( sf )
    {
      dbgprintf ( 5, "sf --> %p\n", sf );
      // compute new directions
      if ( i == 0 )
      {
				new_dir ( sf, esf[1], ndir[i] );
      }
      else
      {
				new_dir ( esf[0], sf, ndir[i] );
      }
      geom::normalize3d ( ndir[i] );
      predicate::is_horizontal(ndir[i]);
			
      if ( ndir[i][2] < 0.0 )
      {
        dbgprintf ( 2, "ndir[2] < 0.0 --> return\n" );
        return;
      }
    }
  }

  se->set_stopped();

  // create new edges
  SweepEdge *elist[2];
  for ( int i = 0; i < 2; i++ )
  {
    SweepFacet *sf = se->get_sweepfacet ( i );
    if ( sf )
    {
      dbgprintf ( 2, "sf --> %p\n", sf );
      dbgprintf ( 5, "dot product --> %lf\n", geom::scalar_product2d ( sf->get_dir2d(), ndir[i] ) );
      dbgprintf ( 5, "plane dir --> %lf %lf\n",  sf->get_dir2d() [0], sf->get_dir2d() [1] );
      dbgprintf ( 2, "new dir --> %.20lf %.20lf %.20lf\n", ndir[i][0], ndir[i][1], ndir[i][2] );

      if ( ndir[i][2] >= 0.0 )
      {
        SweepEdge *nse = new SweepEdge ( se->get_current(), ndir[i], time );

        nse->update_speeds ( sf->get_plane() );
        FrontLine *fl = sf->get_frontline();
        fl->update_sweepedge ( se, nse );

        // add switch event
        Event *nes = econtainer->get();
        nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
        add_priority_event ( nes );
        elist[i] = nse;
      }
    }
  }
  
  if ( esf[0] == esf[1] )
  {
    FrontLine *fl = msf->get_frontline();
    fl->insert_edges ( elist[1], elist[0], pos, time ); // note the change in orientation
  }
  else if ( esf[1] == msf )
  {
    sefr[0]->set_stopped();
    esf[0]->get_frontline()->update_sweepedge ( sefr[0], elist[1] );
    msf->get_frontline()->update_sweepedge ( sefr[0], elist[0] );
  }
  else if ( esf[0] == msf )
  {
    sefr[1]->set_stopped();
    msf->get_frontline()->update_sweepedge ( sefr[1], elist[1] );
    esf[1]->get_frontline()->update_sweepedge ( sefr[1], elist[0] );
  }
}

// void Graph::stop_closest_front ( Front *fr0, Front *fr1, double *ndir )
// {
//   SweepEdge *se = fr0->get_sweepedge ( 1 );
//   ASSERT_ERROR ( fr0->get_sweepedge ( 1 ) == fr1->get_sweepedge ( 0 ), "sweepedges do not correspond\n" );
//   SweepEdge *se0 = fr0->get_sweepedge ( 0 );
//   SweepEdge *se1 = fr1->get_sweepedge ( 1 );
//   double pt0[2];
//   geom::lines2d_intersection ( se->get_start()->p, se0->get_start()->p, se->get_dir2d(), se0->get_dir2d(), Parameters::instance()->vorosweep.epsilon, pt0 );
//   double pt1[2];
//   geom::lines2d_intersection ( se->get_start()->p, se1->get_start()->p, se->get_dir2d(), se1->get_dir2d(), Parameters::instance()->vorosweep.epsilon, pt1 );
//   double d0 = geom::sqrtdistance2d ( se->get_start()->p, pt0 );
//   double d1 = geom::sqrtdistance2d ( se->get_start()->p, pt1 );
// 
//   double time = se->get_current()->p[2];
// 
//   SweepEdge *nse = new SweepEdge ( se->get_current(), ndir, time );
//   nse->set_xyspeed ( INFINITY );
//   FrontLine *fl0 = fr0->get_frontline();
//   fl0->update_sweepedge ( se, nse );
//   FrontLine *fl1 = fr1->get_frontline();
//   fl1->update_sweepedge ( se, nse );
// 
//   if ( d0 < d1 )
//   {
//     se0->merge_current ( se, time );
// // 		se0->set_stopped();
// //     fl0->close_front ( fr0 );
//     merge ( se0, nse );
//   }
//   else
//   {
//     se1->merge_current ( se, time );
// //     se1->set_stopped();
// //     fl1->close_front ( fr1 );
//     merge ( nse, se1 );
//   }
// }


