// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"

using namespace vorosweep;

void Graph::handle_edgeswitch ( Event *ev )
{
  SweepEdge *se = ev->get_sweepedge();
  int gid = ev->get_gid();

  se->print();

  if ( gid == -1 )
  {
    se->set_stopped();
    return;
  }
  
  if ( se->is_stopped() )
    return;

  double *pos = ev->get_coord();
  double time = ev->get_time();

  se->update_current ( time );

  int old = se->get_last_gid();
  bucketgrid->remove_sweepedge ( se, old );
  se->set_last_gid ( gid );

  // clean queue of facets
  for ( int i = 0; i < 2; i++ )
  {
    if ( se->get_sweepfacet ( i ) ) // careful to border edges
    {
      SweepFacet *sf = se->get_sweepfacet ( i );
      sf->remove_earliest_event ( ev );
    }
  }

  insert_edge_in_bucket ( se, pos, time, gid );
}

void Graph::insert_edge_in_bucket ( SweepEdge *se, double *pos, double time, int gid )
{
  // compute next edge switch event
  Event *next = next_edgeswitch ( se, pos, gid );
  se->set_next_switch ( next );
  // add sweepedge to the newly entered cell
  bucketgrid->insert_sweepedge ( se, gid );
  dbgprintf ( 5, "sweepedge se dir --> %lf %lf\n",  se->get_dir2d() [0], se->get_dir2d() [1] );
  
	// check for potential crash events in cell gid
  process_edge_events ( se, time, gid );
	
  // add event for facets
  for ( int i = 0; i < 2; i++ )
  {
    if ( se->get_sweepfacet ( i ) ) // careful to border edges
    {
      SweepFacet *sf = se->get_sweepfacet ( i );
      insert_facet_in_bucket ( sf, pos, time, gid );
    }
  }

}
