// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"
#include "border.h"
#include "sweepobject.h"

using namespace vorosweep;

void Graph::handle_facet_bordercrash ( Event *ev )
{
  SweepFacet *sf = ev->get_sweepfacet();

  Border *bo = ev->get_border();
  bo->print();

  double *pos = ev->get_coord();
  double time = ev->get_time();
  double gid = ev->get_gid();

  int singular[2] = {0, 0};

  FrontLine *fl = sf->get_frontline();
  Front *fr = fl->get_including_front ( pos, time, singular );

  if ( fr == NULL )
  {
    dbgprintf ( 2, "outside front\n" );
    add_earliest_event ( sf );
    return;
  }

  fr->get_sweepedge ( 0 )->print();
  fr->get_sweepedge ( 1 )->print();

  // first identify which point the facet encountered
  int ptindex;
  Border *bo0 = NULL;
  Border *bo1 = NULL;
  if ( geom::sqrtdistance2d ( pos, bo->get_point ( 0 )->p ) < geom::sqrtdistance2d ( pos, bo->get_point ( 1 )->p ) )
  {
    ptindex = 0;
    bo0 = bo;
    bo1 = bo->get_adj ( 1 );
    if ( bo->has_point_been_reached ( 0 ) )
    {
      dbgprintf ( 2, "point 0 of border have already been reached\n" );
      add_earliest_event ( sf );
      return;
    }
  }
  else
  {
    ptindex = 1;
    bo0 = bo->get_adj ( 0 );
    bo1 = bo;
    if ( bo->has_point_been_reached ( 1 ) )
    {
      dbgprintf ( 2, "point 0 of border have already been reached\n" );
      add_earliest_event ( sf );
      return;
    }
  }

  // check if the border is not the same that one of the front edge is following
  if ( fr->get_sweepedge ( 0 )->get_type() == SweepObject::BORDEREDGE )
  {
    Border *cbo = reinterpret_cast<BorderSweepEdge *> ( fr->get_sweepedge ( 0 ) )->get_border();
    cbo->print();
    if ( cbo == bo0 )
    {
      add_earliest_event ( sf );
      return;
    }
  }

  if ( fr->get_sweepedge ( 1 )->get_type() == SweepObject::BORDEREDGE )
  {
    Border *cbo = reinterpret_cast<BorderSweepEdge *> ( fr->get_sweepedge ( 1 ) )->get_border();
    cbo->print();
    if ( cbo == bo1 )
      return;
  }

  // then create new edges
  double ndir[2][3];
  geom::dir2d_to_dir3d ( bo0->get_dir(), sf->get_plane(), ndir[0] );
  geom::dir2d_to_dir3d ( bo1->get_dir(), sf->get_plane(), ndir[1] );

  geom::multvec3d ( ndir[1], -1.0 ); // revert ndir[1]


  double *se0dir = sf->get_angular_bound ( 0 );
  double *se1dir = sf->get_angular_bound ( 1 );

  SweepEdge *nse0 = NULL;
  SweepEdge *nse1 = NULL;

  if ( ptindex == 0 )
  {
    bo->set_point_reached ( 0 );
    Point *pt = new Point;
    pt->set ( bo->get_point ( 0 )->p );
    pt->p[2] = time;
    BorderSweepEdge *nsbe0 = new BorderSweepEdge ( pt, ndir[0], time );
    nsbe0->update_speeds ( sf->get_plane() );
    nsbe0->set_border ( bo0 );

    // add switch event
    Event *nes = econtainer->get();
    nes->init ( nsbe0, bo0, nsbe0->get_start()->p, time, Event::NEWEDGESWITCH, gid );
    add_priority_event ( nes );

    nse0 = nsbe0;

    if ( geom::cross_product2d ( ndir[1], se0dir ) > 0.0 )
    {
      BorderSweepEdge *nsbe1 = new BorderSweepEdge ( pt, ndir[1], time );
      nsbe1->set_border ( bo1 );
      // add switch event
      Event *nes = econtainer->get();
      nes->init ( nsbe1, bo1, nsbe1->get_start()->p, time, Event::NEWEDGESWITCH, gid );
      add_priority_event ( nes );
      nsbe1->update_speeds ( sf->get_plane() );
      nse1 = nsbe1;
    }
    else
    {
      nse1 = singular_point_rotation_ccw ( sf, bo->get_adj ( 1 ), pt, time );
    }
    nse1->update_speeds ( sf->get_plane() );
  }
  else
  {
    bo->set_point_reached ( 1 );
    Point *pt = new Point;
    pt->set ( bo->get_point ( 1 )->p );
    pt->p[2] = time;

    if ( geom::cross_product2d ( se1dir, ndir[0] ) >= 0.0 )
    {
      BorderSweepEdge *nsbe0 = new BorderSweepEdge ( pt, ndir[0], time );
      nsbe0->set_border ( bo0 );
      // add switch event
      Event *nes = econtainer->get();
      nes->init ( nsbe0, bo0, nsbe0->get_start()->p, time, Event::NEWEDGESWITCH, gid );
      add_priority_event ( nes );
      nsbe0->update_speeds ( sf->get_plane() );
      nse0 = nsbe0;
    }
    else
    {
      nse0 = singular_point_rotation_cw ( sf, bo->get_adj ( 0 ), pt, time );
    }

//     printf("ndir --> %lf %lf %lf\n", ndir[1][0], ndir[1][1], ndir[1][2]);
    BorderSweepEdge *nsbe1 = new BorderSweepEdge ( pt, ndir[1], time );
    nsbe1->update_speeds ( sf->get_plane() );
    nsbe1->set_border ( bo1 );
    // add switch event
    Event *nes = econtainer->get();
    nes->init ( nsbe1, bo1, nsbe1->get_start()->p, time, Event::NEWEDGESWITCH, gid );
    add_priority_event ( nes );

    nse1 = nsbe1;
  }

  fl->insert_edges ( nse0, nse1, pos, time ); // note the change in orientation

  add_earliest_event ( sf );
}



// this routine generate as much as facet as needed to circumvent obstacle
SweepEdge *Graph::singular_point_rotation_ccw ( SweepFacet *sf, Border *bo, Point *from, double time )
{
  dbgprintf ( 2, "--> COUNTERCLOCKWISE rotation\n" );
  Generator *g = sf->get_generator();

  int gid = bucketgrid->get_index ( from->p );

  double borderdir[2];
  geom::copyvec2d ( bo->get_dir(), borderdir );
  geom::multvec2d ( borderdir, -1.0 );

  double ndir[3];

  int index = sf->get_pos();
  int nbfacet = 0;

  double *sedir = sf->get_angular_bound ( 0 );
  SweepEdge *lastse = new SweepEdge ( from, sedir, time );
  lastse->update_speeds ( g->get_next_normal ( index ) );
  lastse->set_init();

  SweepEdge *inse1 = lastse;

  // add switch event
  Event *nes = econtainer->get();
  nes->init ( lastse, lastse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
  add_priority_event ( nes );

  double *se0dir = NULL;
  double *se1dir = NULL;

  dbgprintf ( 2, "init normal --> %lf %lf %lf\n", sf->get_plane() [0], sf->get_plane() [1], sf->get_plane() [2] );
  do
  {
    int nextindex = g->get_next_index ( index );
    double *nextnormal = g->get_normal ( nextindex );
    double *nextdir = g->get_dir ( nextindex );
    double nextspeed = g->get_speed ( nextindex );

    dbgprintf ( 2, "checking with normal --> %lf %lf %lf\n", nextnormal[0], nextnormal[1], nextnormal[2] );
    geom::dir2d_to_dir3d ( borderdir, nextnormal, ndir );
//     geom::multvec3d ( ndir, -1.0 );
    dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );

    SweepFacet *nsf = new SweepFacet ( from, nextnormal, nextdir, time );
    SweepEdge *nse;

    se0dir = g->get_edge ( nextindex, 0 );
    se1dir = g->get_edge ( nextindex, 1 );

    dbgprintf ( 2, "border dir --> %lf %lf\n", borderdir [0], borderdir [1] );
    dbgprintf ( 2, "dir 0 --> %lf %lf\n", se0dir[0], se0dir[1] );
    dbgprintf ( 2, "dir 1 --> %lf %lf\n", se1dir[0], se1dir[1] );
    dbgprintf ( 2, "cross --> %lf %lf\n", geom::cross_product2d ( borderdir, se0dir ), geom::cross_product2d ( se1dir, borderdir ) );

    if ( !geom::vector_is_between ( borderdir, se0dir, se1dir ) )
    {
      nse = new SweepEdge ( from, se0dir, time );
      nse->set_init();
    }
    else
    {
      BorderSweepEdge *nbse = new BorderSweepEdge ( from, ndir, time );
      nbse->set_border ( bo );
      nse = nbse;
    }
    // add switch event
    Event *nes = econtainer->get();
    nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
    add_priority_event ( nes );

    nsf->set_pos ( nextindex );
    nsf->set_xyspeed ( nextspeed );
    nsf->set_generator ( g );
    g->add_sweepfacet ( nsf );
    new FrontLine ( nsf, nse, lastse );
    nse->update_speeds ( nsf->get_plane() );

    // add switch event
    Event *nfs = econtainer->get();
    nfs->init ( nsf, nsf->get_start()->p, time, Event::NEWFACETSWITCH, gid );
    add_priority_event ( nfs );

    index = nextindex;
    lastse = nse;

    ASSERT_ERROR ( nbfacet <= g->initial_face_size(), "full rotation without stopping at %lf %lf\n", nse->get_current()->p[0], nse->get_current()->p[1] );

    nbfacet++;
  }
  while ( !geom::vector_is_between ( borderdir, se0dir, se1dir ) );

  return inse1;
}

// this routine generate as much as facet as needed to circumvent obstacle
SweepEdge *Graph::singular_point_rotation_cw ( SweepFacet *sf, Border *bo, Point *from, double time )
{
  dbgprintf ( 2, "--> CLOCKWISE rotation\n" );
//   int bgid = 0;
  Generator *g = sf->get_generator();

  int gid = bucketgrid->get_index ( from->p );

  double borderdir[2];
  geom::copyvec2d ( bo->get_dir(), borderdir );

  double ndir[3];

  int index = sf->get_pos();
  int nbfacet = 0;

  double *sedir = sf->get_angular_bound ( 1 );
  SweepEdge *lastse = new SweepEdge ( from, sedir, time );
  lastse->update_speeds ( g->get_prev_normal ( index ) );
  lastse->set_init();

  SweepEdge *inse0 = lastse;

  // add switch event
  Event *nes = econtainer->get();
  nes->init ( lastse, lastse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
  add_priority_event ( nes );

  double *se0dir = NULL;
  double *se1dir = NULL;

  dbgprintf ( 2, "init normal --> %lf %lf %lf\n", sf->get_plane() [0], sf->get_plane() [1], sf->get_plane() [2] );
  do
  {
    int previndex = g->get_prev_index ( index );
    double *prevnormal = g->get_normal ( previndex );
    double *prevdir = g->get_dir ( previndex );
    double prevspeed = g->get_speed ( previndex );

    dbgprintf ( 2, "checking with normal --> %lf %lf %lf\n", prevnormal[0], prevnormal[1], prevnormal[2] );
    geom::dir2d_to_dir3d ( borderdir, prevnormal, ndir );
    dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );

    SweepFacet *nsf = new SweepFacet ( from, prevnormal, prevdir, time );
    SweepEdge *nse;

    se0dir = g->get_edge ( previndex, 0 );
    se1dir = g->get_edge ( previndex, 1 );

    dbgprintf ( 2, "border dir --> %lf %lf\n", borderdir [0], borderdir [1] );
    dbgprintf ( 2, "dir 0 --> %lf %lf\n", se0dir[0], se0dir[1] );
    dbgprintf ( 2, "dir 1 --> %lf %lf\n", se1dir[0], se1dir[1] );
    dbgprintf ( 2, "cross --> %lf %lf\n", geom::cross_product2d ( borderdir, se0dir ), geom::cross_product2d ( se1dir, borderdir ) );

    if ( !geom::vector_is_between ( borderdir, se0dir, se1dir ) )
    {
      nse = new SweepEdge ( from, se1dir, time );
      nse->set_init();
    }
    else
    {
      BorderSweepEdge *nbse = new BorderSweepEdge ( from, ndir, time );
      nbse->set_border ( bo );
      nse = nbse;
    }

    // add switch event
    Event *nes = econtainer->get();
    nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
    add_priority_event ( nes );

    nsf->set_pos ( previndex );
//     printf("prev speed --> %lf\n", prevsf->get_xyspeed());
    nsf->set_xyspeed ( prevspeed );
    nsf->set_generator ( g );
    g->add_sweepfacet ( nsf );
    new FrontLine ( nsf, lastse, nse );
    nse->update_speeds ( nsf->get_plane() );

    // add switch event
    Event *nfs = econtainer->get();
    nfs->init ( nsf, nsf->get_start()->p, time, Event::NEWFACETSWITCH, gid );
    add_priority_event ( nfs );

    index = previndex;
    lastse = nse;
		
    ASSERT_ERROR ( nbfacet <= g->initial_face_size(), "full rotation without stopping at %lf %lf\n", nse->get_current()->p[0], nse->get_current()->p[1] );

    nbfacet++;
  }
  while ( !geom::vector_is_between ( borderdir, se0dir, se1dir ) );

  return inse0;
}

