// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"

using namespace vorosweep;

// consists only in adding the facet in the buckets
void Graph::handle_facetswitch ( Event *ev )
{
  SweepFacet *sf = ev->get_sweepfacet();
  if ( sf->is_stopped() )
    return;
		sf->print();
  double *pos = ev->get_coord();
  double time = ev->get_time();
  // we need to check if pos is still in the front of the facet
  FrontLine *fl = sf->get_frontline();
  if ( !fl->valid_event ( pos, time ) )
  {

    add_earliest_event ( sf );
    return;
  }

  int gid = ev->get_gid();

  if ( gid == -1 ) // no need to set the facet stopped
    return;

  // add mc to the newly entered cell
	insert_facet_in_bucket(sf, pos, time, gid);
}

void Graph::insert_facet_in_bucket(SweepFacet *sf, double *pos, double time, int gid)
{
// add mc to the newly entered cell
  dbgprintf ( 2, "--> inserting facet in cell %d\n", gid );
  // if insertion failed, facet was already in bucket
  if ( !bucketgrid->insert_sweepfacet ( sf, gid ) )
  {
    add_earliest_event ( sf );
    return;
  }
  // compute next facet switch events
  Event *eventlist[MAX_EVENT_LIST];
  int nbe = 0;
  next_facetswitch ( sf, pos, gid, eventlist, nbe );

  // check for potential crash events in cell gid
  process_facet_events ( sf, time, gid );
}
