// Vorosweep - Copyright (C) 2010-2014 T. Mouton
//
// See the LICENSE.txt file for license information. Please report all
// bugs and problems to <thibaud.mouton@gmail.com>.

#include "vorosweep.h"
#include "border.h"

using namespace vorosweep;

void Graph::merge_and_go_right ( SweepEdge *se0, SweepEdge *se1, SweepFacet *right )
{
  dbgprintf ( 2, "--> right border merging\n" );
  // generate new edge
  double ndir[3];
  right->dir2d_to_dir3d ( se0->get_dir2d(), ndir );
  dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );
  dbgprintf ( 2, "se1 is init --> %d\n", se1->is_init() );

  if ( /*se1->is_init() && */ndir[2] >= 0.0 )
  {
    BorderSweepEdge *bse0 = static_cast<BorderSweepEdge *> ( se0 );
    Border *bo = bse0->get_border();
//     Grid1d *bordergrid = bo->get_grid();
    dbgprintf ( 2, "--> creating new edge from edges %p %p with facet %p :\n", se0, se1, right );
    BorderSweepEdge *nse = new BorderSweepEdge ( se1->get_current(), ndir, get_time() );
    nse->set_border ( bo );
//     nse->set_generator ( right->get_generator() );
    nse->update_speeds ( right->get_plane() );
    FrontLine *fl1 = right->get_frontline();
    fl1->update_sweepedge ( se1, nse );
//     int bgid = bordergrid->get_index ( nse->get_start()->p );
//     Event *ne = econtainer->get();
//     ne->init ( nse, bo, nse->get_start()->p, get_time(), Event::NEWBORDEREDGESWITCH, bgid );
//     add_event ( ne );
//     // add also switch event in grid for facet switch
//     int gid = bucketgrid->get_index ( nse->get_start()->p );
//     Event *next = next_edgeswitch ( nse, nse->get_start()->p, gid );
//     dbgprintf ( 5, "--> next switch event add :\n" );
//     add_event ( next );

    int gid = bucketgrid->get_index ( nse->get_start()->p );
    Event *ne = econtainer->get();
    ne->init ( nse, nse->get_start()->p, get_time(), Event::NEWEDGESWITCH, gid );
    add_priority_event ( ne );
  }
  else
  {
    Front *fr1 = se1->get_front ( 1 );
    SweepEdge *seopp = fr1->get_sweepedge ( 1 );
    se1->merge_current ( seopp, get_time() );
    FrontLine *fl1 = fr1->get_frontline();
    fl1->close_front ( fr1 );
    seopp->set_stopped();
  }
}

void Graph::merge_and_go_left ( SweepEdge *se0, SweepEdge *se1, SweepFacet *left )
{
  dbgprintf ( 2, "--> left border merging\n" );
  // generate new edge
  double ndir[3];
  left->dir2d_to_dir3d ( se1->get_dir2d(), ndir );
  dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );
  dbgprintf ( 2, "se0 is init --> %d\n", se0->is_init() );

  if ( /*se0->is_init() && */ndir[2] >= 0.0 )
  {
    BorderSweepEdge *bse1 = static_cast<BorderSweepEdge *> ( se1 );
    Border *bo = bse1->get_border();
//     Grid1d *bordergrid = bo->get_grid();
    dbgprintf ( 2, "--> creating new edge from edges %p %p with facet %p :\n", se0, se1, left );
    BorderSweepEdge *nse = new BorderSweepEdge ( se0->get_current(), ndir, get_time() );
    nse->set_border ( bo );
//     nse->set_generator ( left->get_generator() );
    nse->update_speeds ( left->get_plane() );
    FrontLine *fl0 = left->get_frontline();
    fl0->update_sweepedge ( se0, nse );
//     int bgid = bordergrid->get_index ( nse->get_start()->p );
//     Event *ne = econtainer->get();
//     ne->init ( nse, bo, nse->get_start()->p, get_time(), Event::NEWBORDEREDGESWITCH, bgid );
//     add_event ( ne );
//     // add also switch event in grid for facet switch
//     int gid = bucketgrid->get_index ( nse->get_start()->p );
//     Event *next = next_edgeswitch ( nse, nse->get_start()->p, gid );
//     dbgprintf ( 5, "--> next switch event add :\n" );
//     add_event ( next );

    int gid = bucketgrid->get_index ( nse->get_start()->p );
    Event *ne = econtainer->get();
    ne->init ( nse, nse->get_start()->p, get_time(), Event::NEWEDGESWITCH, gid );
    add_priority_event ( ne );
  }
  else
  {
    Front *fr0 = se0->get_front ( 0 );
    SweepEdge *seopp = fr0->get_sweepedge ( 0 );
    se0->merge_current ( seopp, get_time() );
    // close the front
    FrontLine *fl0 = fr0->get_frontline();
    fl0->close_front ( fr0 );
    seopp->set_stopped();
  }
}

// this function MUST be cleaned !!
void Graph::voronoi_vertex_merge ( SweepEdge *se0, SweepEdge *se1, double *ndir )
{
  // get the other edge
  SweepEdge *edge[2];

  Front *fr0 = se0->get_front ( 0 );
  Front *fr1 = se1->get_front ( 1 );

  edge[0] = fr0->get_sweepedge ( 0 );
  edge[1] = fr1->get_sweepedge ( 1 );

  if ( edge[0] == edge[1] ) // 3 point intersection case
  {
    dbgprintf ( 2, "--> stopping everything (3 edges)\n" );
    fr0->get_frontline()->close_front ( fr0 );
    fr1->get_frontline()->close_front ( fr1 );
    se0->merge_current ( edge[0], get_time() );
    edge[0]->set_stopped();
  }
  else // exotic cases
  {
    dbgprintf ( 2, "msf --> %p %p\n", se0->get_sweepfacet ( 0 ), se1->get_sweepfacet ( 1 ) );
//     dbgprintf ( 2, "4 edge cases !!\n" );

    // extedge contains the extreme edge that meet at the same point as se0 and se1
    SweepEdge *extedge[2];
    extedge[0] = se0;
    extedge[1] = se1;

    // ultimate test --> get the fronts
    Front *fr0 = extedge[0]->get_front ( 0 );
    Front *fr1 = extedge[1]->get_front ( 1 );

    // case we merge 2 front of the same frontline
    if ( fr0 && fr1 && fr0->get_owner() == fr1->get_owner() )
    {
      FrontLine *fl0 = fr0->get_frontline();
      FrontLine *fl1 = fr1->get_frontline();

      ASSERT_ERROR ( fl0 == fl1, "merging different frontline : %p %p\n", fl0, fl1 );

      Front *frn = fl0->front_merge ( fr0, fr1 );
      // generate the new edgecrash
      SweepEdge *frnse0 = frn->get_sweepedge ( 0 );
      SweepEdge *frnse1 = frn->get_sweepedge ( 1 );
      next_edgecrash ( frnse0, frnse1 );
    }
    else if ( fr0->get_owner()->get_generator() != fr1->get_owner()->get_generator() )
    {
      // we create 2 new edges
      new_facet ( se0, se1, ndir );
    }
  }
}

void Graph::new_facet ( SweepEdge *se0, SweepEdge *se1, double *ndir )
{
  se0->print();
  se1->print();

  SweepFacet *sf0 = se0->get_sweepfacet ( 0 );
  SweepFacet *sf1 = se1->get_sweepfacet ( 1 );

  sf0->print();
  sf1->print();

  // CounterClockWise
  if ( geom::cross_product2d ( sf1->get_angular_bound ( 0 ), ndir ) > 0.0 )
  {
    rotation_ccw ( se0, se1 );
  }
  else
  {
    rotation_cw ( se0, se1 );
  }
}

// this routine generate as much as facet as needed to circumvent obstacle
void Graph::rotation_ccw ( SweepEdge *see0, SweepEdge *see1 )
{
  dbgprintf ( 2, "--> COUNTERCLOCKWISE rotation\n" );
  SweepFacet *sf0 = see0->get_front ( 0 )->get_owner();
  SweepFacet *sf1 = see1->get_front ( 1 )->get_owner();
  Generator *g = sf1->get_generator();

  double time = get_time();
  int gid = bucketgrid->get_index ( see1->get_current()->p );

  int index = sf1->get_pos();
  //   double *normal = sf1->get_plane();
  double *normal = g->get_normal ( index );

  SweepEdge *lastse = NULL;
  double *se0dir = sf1->get_angular_bound ( 0 );
  double *se1dir = sf1->get_angular_bound ( 1 );

  // compute direction
  double ndir[3];
  geom::cross_product3d ( sf0->get_plane(), normal, ndir );
  dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );

//   if ( geom::cross_product2d ( ndir, se0dir ) < 0.0 || geom::cross_product2d ( se1dir, ndir ) < 0.0 )
  if ( !geom::vector_is_between ( ndir, se0dir , se1dir ) )
  {
    lastse = new SweepEdge ( see1->get_current(), se0dir, time );
    dbgprintf ( 2, "dir edge --> %lf %lf %lf\n",  se0dir [0], se0dir [1], se0dir [2] );
    lastse->update_speeds ( g->get_next_normal ( index ) );
    lastse->set_init();
  }
  else
  {
    lastse = new SweepEdge ( see1->get_current(), ndir, time );
    dbgprintf ( 2, "dir edge --> %lf %lf %lf\n",  ndir[0], ndir[1], ndir[2] );
    lastse->update_speeds ( sf1->get_plane() );
  }

  // stopping see1
  see1->set_stopped();
  see1->get_front ( 1 )->get_frontline()->update_sweepedge ( see1, lastse );

  // add switch event
  Event *nes = econtainer->get();
  nes->init ( lastse, lastse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
  add_priority_event ( nes );

  dbgprintf ( 2, "init normal --> %lf %lf %lf\n", sf1->get_plane() [0], sf1->get_plane() [1], sf1->get_plane() [2] );

  int count = 0;
  while ( !geom::vector_is_between ( ndir, se0dir , se1dir ) )
  {
    int nextindex = g->get_next_index ( index );
    double *nextnormal = g->get_normal ( nextindex );
    double *nextdir = g->get_dir ( nextindex );
    double nextspeed = g->get_speed ( nextindex );

    dbgprintf ( 2, "checking with normal --> %lf %lf %lf\n", nextnormal[0], nextnormal[1], nextnormal[2] );

    geom::cross_product3d ( sf0->get_plane(), nextnormal, ndir );
    dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );

    SweepFacet *nsf = new SweepFacet ( see1->get_current(), nextnormal, nextdir, time );
    SweepEdge *nse;

    se0dir = g->get_edge ( nextindex, 0 );
    se1dir = g->get_edge ( nextindex, 1 );

//     printf ( "dir 0 --> %lf %lf\n", se0dir [0], se0dir [1] );
//     printf ( "dir 1 --> %lf %lf\n", se1dir [0], se1dir [1] );
    if ( !geom::vector_is_between ( ndir, se0dir , se1dir ) )
      nse = new SweepEdge ( see1->get_current(), se0dir, time );
    else
      nse = new SweepEdge ( see1->get_current(), ndir, time );
    nse->set_init();
    // add switch event
    Event *nes = econtainer->get();
    nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
    add_priority_event ( nes );

    nsf->set_pos ( nextindex );
    nsf->set_xyspeed ( nextspeed );
    nsf->set_generator ( g );
    g->add_sweepfacet ( nsf );
    new FrontLine ( nsf, nse, lastse );
    nse->update_speeds ( nsf->get_plane() );

    // add switch event
    Event *nfs = econtainer->get();
    nfs->init ( nsf, nsf->get_start()->p, time, Event::NEWFACETSWITCH, gid );
    add_priority_event ( nfs );

    index = nextindex;
    lastse = nse;

    ASSERT_ERROR ( count <= g->face_size(), "full rotation without stopping at %lf %lf\n", see0->get_current()->p[0], see0->get_current()->p[1] );

// 		printf("count --> %d\n", count);
    count++;
  }

  // stopping see1
  see0->set_stopped();
  see0->get_front ( 0 )->get_frontline()->update_sweepedge ( see0, lastse );
}

// this routine generate as much as facet as needed to circumvent obstacle
void Graph::rotation_cw ( SweepEdge *see0, SweepEdge *see1 )
{
  dbgprintf ( 2, "--> CLOCKWISE rotation\n" );
  SweepFacet *sf0 = see0->get_front ( 0 )->get_owner();
  SweepFacet *sf1 = see1->get_front ( 1 )->get_owner();
  Generator *g = sf0->get_generator();

  if ( g == sf1->get_generator() )
  {
    printf ( "rotation not allowed\n" );
    return;
  }

  double time = get_time();
  int gid = bucketgrid->get_index ( see0->get_current()->p );

  double *normal = sf0->get_plane();
  double ndir[3];

  int index = sf0->get_pos();

  dbgprintf ( 2, "index --> %d\n", index );
  dbgprintf ( 2, "normal --> %lf %lf\n", normal[0], normal[1] );

  SweepEdge *lastse = NULL;
  double *se0dir = sf0->get_angular_bound ( 0 );
  double *se1dir = sf0->get_angular_bound ( 1 );

  // compute direction
  geom::cross_product3d ( normal, sf1->get_plane(), ndir );
  dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );
  dbgprintf ( 2, "cross --> %lf %lf\n", geom::cross_product2d ( ndir, se0dir ), geom::cross_product2d ( se1dir, ndir ) );

  if ( !geom::vector_is_between ( ndir, se0dir , se1dir ) )
  {
    lastse = new SweepEdge ( see0->get_current(), se1dir, time );
    dbgprintf ( 2, "dir edge --> %lf %lf %lf\n",  se1dir [0], se1dir [1], se1dir [2] );
    lastse->update_speeds ( g->get_prev_normal ( index ) );
    lastse->set_init();
  }
  else
  {
    lastse = new SweepEdge ( see1->get_current(), ndir, time );
    dbgprintf ( 2, "dir edge --> %lf %lf %lf\n",  ndir[0], ndir[1], ndir[2] );
    lastse->update_speeds ( sf0->get_plane() );
  }

  // stopping see0
  see0->set_stopped();
  see0->get_front ( 0 )->get_frontline()->update_sweepedge ( see0, lastse );

  // add switch event
  Event *nes = econtainer->get();
  nes->init ( lastse, lastse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
  add_priority_event ( nes );

  dbgprintf ( 2, "init normal --> %lf %lf %lf\n", sf0->get_plane() [0], sf0->get_plane() [1], sf0->get_plane() [2] );
  dbgprintf ( 2, "compare with normal --> %lf %lf %lf\n", sf1->get_plane() [0], sf1->get_plane() [1], sf1->get_plane() [2] );
  int count = 0;

  while ( !geom::vector_is_between ( ndir, se0dir , se1dir ) )
  {
    int previndex = g->get_prev_index ( index );
    double *prevnormal = g->get_normal ( previndex );
    double *prevdir = g->get_dir ( previndex );
    double prevspeed = g->get_speed ( previndex );

    dbgprintf ( 2, "checking with normal --> %lf %lf %lf\n", prevnormal[0], prevnormal[1], prevnormal[2] );

    geom::cross_product3d ( prevnormal, sf1->get_plane(), ndir );
    dbgprintf ( 2, "ndir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );

    SweepFacet *nsf = new SweepFacet ( see0->get_current(), prevnormal, prevdir, time );
    SweepEdge *nse;

    se0dir = g->get_edge ( previndex, 0 );
    se1dir = g->get_edge ( previndex, 1 );

//     printf ( "dir 0 --> %lf %lf\n", se0dir [0], se0dir [1] );
//     printf ( "dir 1 --> %lf %lf\n", se1dir [0], se1dir [1] );

    if ( !geom::vector_is_between ( ndir, se0dir , se1dir ) )
      nse = new SweepEdge ( see0->get_current(), se1dir, time );
    else
      nse = new SweepEdge ( see0->get_current(), ndir, time );

    nse->set_init();
    // add switch event
    Event *nes = econtainer->get();
    nes->init ( nse, nse->get_start()->p, time, Event::NEWEDGESWITCH, gid );
    add_priority_event ( nes );

    nsf->set_pos ( previndex );
//     printf ( "prev speed --> %lf\n", prevsf->get_xyspeed() );
    nsf->set_xyspeed ( prevspeed );
    nsf->set_generator ( g );
    g->add_sweepfacet ( nsf );
    new FrontLine ( nsf, lastse, nse );
    nse->update_speeds ( nsf->get_plane() );

    // add switch event
    Event *nfs = econtainer->get();
    nfs->init ( nsf, nsf->get_start()->p, time, Event::NEWFACETSWITCH, gid );
    add_priority_event ( nfs );

    index = previndex;
    lastse = nse;

    ASSERT_ERROR ( count <= g->initial_face_size(), "full rotation without stopping at %lf %lf\n", see0->get_current()->p[0], see0->get_current()->p[1] );

// 		printf("count --> %d\n", count);
    count++;
  }

  // stopping see1
  see1->set_stopped();
  see1->get_front ( 1 )->get_frontline()->update_sweepedge ( see1, lastse );
}


// close cells on a single facet
int Graph::enclose_island ( SweepEdge *se0, SweepEdge *se1 )
{
  if ( se1->get_front ( 1 ) && se0->get_sweepfacet ( 0 ) && se0->get_sweepfacet ( 0 ) == se1->get_front ( 1 )->get_sweepedge ( 1 )->get_sweepfacet ( 1 ) )
  {
    dbgprintf ( 1, "simple closing case 0\n" );
    Front *frint0 = se1->get_front ( 0 );
    Front *frint1 = se1->get_front ( 1 );
    if ( frint0 && frint1 )
    {
      FrontLine *flint0 = frint0->get_frontline();
      FrontLine *flint1 = frint1->get_frontline();
      // close the interior fronts
      flint0->close_front ( frint0 );
      flint1->close_front ( frint1 );

      SweepEdge *se2 = frint1->get_sweepedge ( 1 );
      Front *frext0 = se0->get_front ( 0 );
      Front *frext1 = se2->get_front ( 1 );

      se1->merge_current ( se2, get_time() );
      if ( frext0 && frext1 )
      {
        FrontLine *mfl = frext0->get_frontline();
        Front *frn = mfl->front_merge ( frext0, frext1 );
        // generate the new edgecrash
        SweepEdge *frnse0 = frn->get_sweepedge ( 0 );
        SweepEdge *frnse1 = frn->get_sweepedge ( 1 );
        next_edgecrash ( frnse0, frnse1 );
        return 1;
      }
    }
  }
  if ( se0->get_front ( 0 ) && se1->get_sweepfacet ( 1 ) && se1->get_sweepfacet ( 1 ) == se0->get_front ( 0 )->get_sweepedge ( 0 )->get_sweepfacet ( 0 ) )
  {
    dbgprintf ( 1, "simple closing case 1\n" );
    Front *frint0 = se0->get_front ( 0 );
    Front *frint1 = se0->get_front ( 1 );
    if ( frint0 && frint1 )
    {
      FrontLine *flint0 = frint0->get_frontline();
      FrontLine *flint1 = frint1->get_frontline();
      // close the interior fronts
      flint0->close_front ( frint0 );
      flint1->close_front ( frint1 );

      SweepEdge *se2 = frint0->get_sweepedge ( 0 );
      Front *frext0 = se2->get_front ( 0 );
      Front *frext1 = se1->get_front ( 1 );

      se0->merge_current ( se2, get_time() );
      if ( frext0 && frext1 )
      {
        FrontLine *mfl = frext0->get_frontline();
        Front *frn = mfl->front_merge ( frext0, frext1 );
        // generate the new edgecrash
        SweepEdge *frnse0 = frn->get_sweepedge ( 0 );
        SweepEdge *frnse1 = frn->get_sweepedge ( 1 );
        next_edgecrash ( frnse0, frnse1 );
        return 1;
      }
    }
  }
  return 0;
}

void Graph::new_dir ( double *n0, double *n1, double *ndir )
{
  dbgprintf ( 2, "new dir --> %lf %lf %lf cross %lf %lf %lf\n", n0[0], n0[1], n0[2], n1[0], n1[1], n1[2] );
  geom::cross_product3d ( n0, n1, ndir );
  if ( predicate::null_length ( ndir ) )
    geom::dir2d_to_dir3d ( n0, n0, ndir );
  predicate::is_horizontal ( ndir );
  dbgprintf ( 2, "new dir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );
}

void Graph::new_dir ( SweepFacet *sf0, SweepFacet *sf1, double *ndir )
{
  dbgprintf ( 2, "new dir --> %lf %lf %lf cross %lf %lf %lf\n", sf0->get_plane() [0], sf0->get_plane() [1], sf0->get_plane() [2], sf1->get_plane() [0], sf1->get_plane() [1], sf1->get_plane() [2] );
  geom::cross_product3d ( sf0->get_plane(), sf1->get_plane(), ndir );
  if ( predicate::null_length ( ndir ) )
    geom::dir2d_to_dir3d ( sf0->get_plane(), sf0->get_plane(), ndir );
  predicate::is_horizontal ( ndir );
  dbgprintf ( 2, "new dir --> %lf %lf %lf\n", ndir[0], ndir[1], ndir[2] );
}

SweepEdge *Graph::already_merged ( SweepEdge *se0, SweepEdge *se1 )
{
  if ( !se0->get_front ( 0 ) || !se1->get_front ( 1 ) )
    return NULL;

  return already_merged ( se0->get_front ( 0 ), se1->get_front ( 1 ) );

//   if ( se0->get_front ( 0 )->get_sweepedge ( 0 ) == se1->get_front ( 1 )->get_sweepedge ( 1 ) )
//   {
//     dbgprintf (1, "already merged : continuing ...\n" );
//     return se0->get_front ( 0 )->get_sweepedge ( 0 );
//   }
//   return NULL;
}

SweepEdge *Graph::already_merged ( Front *fr0, Front *fr1 )
{
  if ( fr0->get_sweepedge ( 0 ) == fr1->get_sweepedge ( 1 ) )
  {
    dbgprintf ( 1, "already merged : continuing ...\n" );
    return fr0->get_sweepedge ( 0 );
  }
  return NULL;
}

void Graph::simple_newedgeswitch ( Point *current, SweepEdge *se0, SweepEdge *se1, double *ndir3d, SweepFacet *msf[2] )
{
  dbgprintf ( 2, "--> creating new edge from edges %p %p with facet %p %p :\n", se0, se1, msf[0], msf[1] );
  geom::normalize3d ( ndir3d );
  SweepEdge *nse = new SweepEdge ( current, ndir3d, get_time() );

  nse->update_speeds ( msf[0]->get_plane() );
  FrontLine *fl0 = msf[0]->get_frontline();
  FrontLine *fl1 = msf[1]->get_frontline();

  fl0->update_sweepedge ( se0, nse );
  fl1->update_sweepedge ( se1, nse );

  int gid = se0->get_last_gid();
  if ( se0->get_type() == SweepObject::BORDEREDGE )
    gid = bucketgrid->get_index ( current->p );

  Event *ne = econtainer->get();
  ne->init ( nse, nse->get_start()->p, get_time(), Event::NEWEDGESWITCH, gid );
  add_priority_event ( ne );
}

void Graph::new_edge ( SweepEdge *se0, SweepEdge *se1, double *ndir3d )
{
  SweepFacet *msf[2] = {NULL, NULL};
  msf[0] = se0->get_sweepfacet ( 0 );
  msf[1] = se1->get_sweepfacet ( 1 );
  if ( ndir3d[2] < 0.0 )
  {
    voronoi_vertex_merge ( se0, se1, ndir3d );
  }
  else // ndir[2] >= 0.0
  {
    msf[0]->print();
    msf[1]->print();
// 		double ndir2d[2];
// 		geom::copyvec2d ( ndir3d, ndir2d );
//     geom::normalize2d ( ndir2d );
//     dbgprintf ( 2, "new dir 2d --> %lf %lf\n", ndir2d[0], ndir2d[1] );
    geom::normalize3d ( ndir3d );
    dbgprintf ( 2, "new dir 3d --> %lf %lf %lf\n", ndir3d[0], ndir3d[1], ndir3d[2] );

    Generator *g0 = msf[0]->get_generator();
    Generator *g1 = msf[1]->get_generator();

    // case where the same generator is enclosing one or more cells
    if ( g0 == g1 )
    {
      simple_newedgeswitch ( se0->get_current(), se0, se1, ndir3d, msf );
      return;
    }

    double *ise0dir[2];
    ise0dir[0] = msf[0]->get_angular_bound ( 0 );
    ise0dir[1] = msf[0]->get_angular_bound ( 1 );
    double *ise1dir[2];
    ise1dir[0] = msf[1]->get_angular_bound ( 0 );
    ise1dir[1] = msf[1]->get_angular_bound ( 1 );

    dbgprintf ( 2, "--> is %lf %lf between %lf %lf and %lf %lf\n", ndir3d[0], ndir3d[1], ise1dir[0][0], ise1dir[0][1], ise0dir[1][0], ise0dir[1][1] );

//     if ( geom::vector_is_between ( ndir2d, ise1dir[0], ise0dir[1] ) )
    if ( geom::vector_is_between ( ndir3d, ise1dir[0], ise0dir[1] ) )
    {
      simple_newedgeswitch ( se0->get_current(), se0, se1, ndir3d, msf );
    }
    else
    {
      voronoi_vertex_merge ( se0, se1, ndir3d );
    }
  }
}

// we assume that sweepedges are sorted
void Graph::merge ( SweepEdge *se0, SweepEdge *se1 )
{
  dbgprintf ( 2, "merging edge %lf %lf and %lf %lf at %lf %lf\n", se0->get_dir2d() [0], se0->get_dir2d() [1], se1->get_dir2d() [0], se1->get_dir2d() [1], se1->get_current()->p[0], se1->get_current()->p[1] );

  SweepFacet *msf[2] = {NULL, NULL};
  msf[0] = se0->get_sweepfacet ( 0 );
  msf[1] = se1->get_sweepfacet ( 1 );

  if ( se0->get_front ( 1 ) && se1->get_front ( 0 ) && ( se0->get_front ( 1 ) == se1->get_front ( 0 ) ) )
  {
    Front *cfr = se0->get_front ( 1 );
    dbgprintf ( 2, "common sweepfacet dir --> %lf %lf\n", cfr->get_owner()->get_dir2d() [0], cfr->get_owner()->get_dir2d() [1] );
    // close the front
    FrontLine *cfl = cfr->get_frontline();
    cfl->close_front ( cfr );
  }

  if ( enclose_island ( se0, se1 ) )
  {
    return;
  }

  // create a new edge if necessary
  if ( msf[0] && msf[1] )
  {
    if ( msf[0] == msf[1] ) // corner
    {
      Front *cfr = se0->get_front ( 0 );
      // close the front
      FrontLine *cfl = cfr->get_frontline();
      cfl->close_front ( cfr );
      return;
    }

    if ( already_merged ( se0, se1 ) )
      return;

    // compute new directions
    double ndir3d[3];
    new_dir ( msf[0], msf[1], ndir3d );

//     double ndir2d[2];
//     geom::copyvec2d ( ndir3d, ndir2d );

//     if ( ndir3d[2] < 0.0 )
//     {
//       voronoi_vertex_merge ( se0, se1, ndir3d );
//     }
//     else // ndir[2] >= 0.0
//     {
//       msf[0]->print();
//       msf[1]->print();
//       geom::normalize2d ( ndir2d );
//       dbgprintf ( 2, "new dir 2d --> %lf %lf\n", ndir2d[0], ndir2d[1] );
//
//       Generator *g0 = msf[0]->get_generator();
//       Generator *g1 = msf[1]->get_generator();
//
//       // case where the same generator is enclosing one or more cells
//       if ( g0 == g1 )
//       {
//         simple_newedgeswitch ( se0->get_current(), se0, se1, ndir3d, msf );
//         return;
//       }
//
//       double *ise0dir[2];
//       ise0dir[0] = msf[0]->get_angular_bound ( 0 );
//       ise0dir[1] = msf[0]->get_angular_bound ( 1 );
//       double *ise1dir[2];
//       ise1dir[0] = msf[1]->get_angular_bound ( 0 );
//       ise1dir[1] = msf[1]->get_angular_bound ( 1 );
//
// //       if ( ( geom::cross_product2d ( ise0dir[1], ndir2d ) >= 0.0 ) && geom::cross_product2d ( ise1dir[0], ndir2d ) <= 0.0 )
//       if ( geom::vector_is_between ( ndir2d, ise1dir[0], ise0dir[1] ) )
//       {
//         simple_newedgeswitch ( se0->get_current(), se0, se1, ndir3d, msf );
//       }
//       else
//       {
//         voronoi_vertex_merge ( se0, se1, ndir3d );
//       }
//     }
    new_edge ( se0, se1, ndir3d );
  }
  else if ( msf[0] && !msf[1] ) // relative to border crashes
  {
    merge_and_go_left ( se0, se1, msf[0] );
  }
  else if ( !msf[0] && msf[1] )
  {
    merge_and_go_right ( se0, se1, msf[1] );
  }
  else // if ( !msf[0] && !msf[1] )
  {
    Front *cfr0 = se0->get_front ( 1 );
    cfr0->get_sweepedge ( 1 )->set_stopped();
    FrontLine *cfl0 = cfr0->get_frontline();
    cfl0->close_front ( cfr0 );
    Front *cfr1 = se1->get_front ( 0 );
    cfr1->get_sweepedge ( 0 )->set_stopped();
    FrontLine *cfl1 = cfr1->get_frontline();
    cfl1->close_front ( cfr1 );
  }
}



